import typing as t
import pyarrow as pa

from sarus_data_spec.constants import DATA


def pandas_index_columns(schema: pa.Schema) -> t.List[str]:
    """Return the list of columns that have to be considered as Pandas index
    columns and ignored by the Sarus type.
    """
    pandas_metadata = schema.pandas_metadata
    if pandas_metadata is None:
        return []

    def column_name(index: t.Any) -> t.Optional[str]:
        if isinstance(index, str):
            return index
        elif isinstance(index, dict):
            return t.cast(t.Optional[str], index["name"])
        else:
            raise ValueError("Unrecognized Arrow `index_column` format")

    columns = [
        column_name(index) for index in pandas_metadata["index_columns"]
    ]
    return [col for col in columns if col is not None]


def remove_pandas_index_columns(table: pa.Table) -> pa.Table:
    """Remove pandas metadata and drop additional
    index columns used for Pandas indexing.
    """
    index_columns_names = pandas_index_columns(table.schema)
    return table.drop(index_columns_names).replace_schema_metadata(None)


def convert_pandas_metadata_to_admin_columns(table: pa.Table) -> pa.Table:
    """Isolate the pandas index from the data."""
    index_columns = pandas_index_columns(table.schema)
    if len(index_columns) == 0:
        return table

    # Create admin columns
    data_columns = [
        col for col in table.column_names if col not in index_columns
    ]

    data_arrays = [
        chunked_array.combine_chunks()
        for name, chunked_array in zip(table.column_names, table.columns)
        if name in data_columns
    ]
    index_arrays = [
        chunked_array.combine_chunks()
        for name, chunked_array in zip(table.column_names, table.columns)
        if name in index_columns
    ]

    data_array = pa.StructArray.from_arrays(data_arrays, names=data_columns)

    new_table = pa.Table.from_arrays(index_arrays, names=index_columns)
    new_table = new_table.append_column(DATA, data_array)
    return new_table.replace_schema_metadata(table.schema.metadata)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sarus_data_spec/protobuf/dataset.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n&sarus_data_spec/protobuf/dataset.proto\x12\x0fsarus_data_spec\"\x8c\x08\n\x07\x44\x61taset\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0b\n\x03\x64oc\x18\x03 \x01(\t\x12+\n\x04spec\x18\x04 \x01(\x0b\x32\x1d.sarus_data_spec.Dataset.Spec\x12<\n\nproperties\x18\x05 \x03(\x0b\x32(.sarus_data_spec.Dataset.PropertiesEntry\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\xca\x02\n\x04Spec\x12;\n\x0btransformed\x18\x01 \x01(\x0b\x32$.sarus_data_spec.Dataset.TransformedH\x00\x12-\n\x04\x66ile\x18\x02 \x01(\x0b\x32\x1d.sarus_data_spec.Dataset.FileH\x00\x12/\n\x05\x66iles\x18\x03 \x01(\x0b\x32\x1e.sarus_data_spec.Dataset.FilesH\x00\x12\x33\n\x07\x61rchive\x18\x04 \x01(\x0b\x32 .sarus_data_spec.Dataset.ArchiveH\x00\x12+\n\x03sql\x18\x05 \x01(\x0b\x32\x1c.sarus_data_spec.Dataset.SqlH\x00\x12;\n\x0bhuggingface\x18\x06 \x01(\x0b\x32$.sarus_data_spec.Dataset.HuggingfaceH\x00\x42\x06\n\x04spec\x1a\xbd\x01\n\x0bTransformed\x12\x11\n\ttransform\x18\x01 \x01(\t\x12\x11\n\targuments\x18\x02 \x03(\t\x12Q\n\x0fnamed_arguments\x18\x03 \x03(\x0b\x32\x38.sarus_data_spec.Dataset.Transformed.NamedArgumentsEntry\x1a\x35\n\x13NamedArgumentsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a#\n\x04\x46ile\x12\x0e\n\x06\x66ormat\x18\x01 \x01(\t\x12\x0b\n\x03uri\x18\x02 \x01(\t\x1a,\n\x05\x46iles\x12\x0e\n\x06\x66ormat\x18\x01 \x01(\t\x12\x13\n\x0buri_pattern\x18\x02 \x01(\t\x1a>\n\x07\x41rchive\x12\x0e\n\x06\x66ormat\x18\x01 \x01(\t\x12\x16\n\x0e\x61rchive_format\x18\x02 \x01(\t\x12\x0b\n\x03uri\x18\x03 \x01(\t\x1an\n\x03Sql\x12\x0b\n\x03uri\x18\x01 \x01(\t\x12\x32\n\x06tables\x18\x02 \x03(\x0b\x32\".sarus_data_spec.Dataset.Sql.Table\x1a&\n\x05Table\x12\x0e\n\x06schema\x18\x01 \x01(\t\x12\r\n\x05table\x18\x02 \x01(\t\x1a*\n\x0bHuggingface\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05split\x18\x02 \x01(\tb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'sarus_data_spec.protobuf.dataset_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _DATASET_PROPERTIESENTRY._options = None
  _DATASET_PROPERTIESENTRY._serialized_options = b'8\001'
  _DATASET_TRANSFORMED_NAMEDARGUMENTSENTRY._options = None
  _DATASET_TRANSFORMED_NAMEDARGUMENTSENTRY._serialized_options = b'8\001'
  _DATASET._serialized_start=60
  _DATASET._serialized_end=1096
  _DATASET_PROPERTIESENTRY._serialized_start=219
  _DATASET_PROPERTIESENTRY._serialized_end=268
  _DATASET_SPEC._serialized_start=271
  _DATASET_SPEC._serialized_end=601
  _DATASET_TRANSFORMED._serialized_start=604
  _DATASET_TRANSFORMED._serialized_end=793
  _DATASET_TRANSFORMED_NAMEDARGUMENTSENTRY._serialized_start=740
  _DATASET_TRANSFORMED_NAMEDARGUMENTSENTRY._serialized_end=793
  _DATASET_FILE._serialized_start=795
  _DATASET_FILE._serialized_end=830
  _DATASET_FILES._serialized_start=832
  _DATASET_FILES._serialized_end=876
  _DATASET_ARCHIVE._serialized_start=878
  _DATASET_ARCHIVE._serialized_end=940
  _DATASET_SQL._serialized_start=942
  _DATASET_SQL._serialized_end=1052
  _DATASET_SQL_TABLE._serialized_start=1014
  _DATASET_SQL_TABLE._serialized_end=1052
  _DATASET_HUGGINGFACE._serialized_start=1054
  _DATASET_HUGGINGFACE._serialized_end=1096
# @@protoc_insertion_point(module_scope)



from math import ceil
import logging
import typing as t

import numpy as np
import pandas as pd

from sarus_data_spec.constants import MAX_MAX_MULT
from sarus_data_spec.dataspec_validator.signature import SarusBoundSignature
from sarus_data_spec.protobuf.utilities import unwrap
import sarus_data_spec.typing as st

from ..external_op import (
    DEFAULT_MAX_MAX_MULT,
    ExternalOpImplementation,
    retrieve_max_mult_info_from_parent,
)

logger = logging.getLogger(__name__)

try:
    from sarus_data_spec.sarus_statistics.ops.bounds.op import BoundOp
    from sarus_data_spec.sarus_statistics.ops.corr.op import CorrOp
    from sarus_data_spec.sarus_statistics.ops.histograms.op import (
        CountOp,
        HistogramOp,
    )
    from sarus_data_spec.sarus_statistics.ops.max_multiplicity.op import (
        MaxMultiplicityOp,
    )
    from sarus_data_spec.sarus_statistics.ops.mean.op import MeanOp
    from sarus_data_spec.sarus_statistics.ops.median.op import MedianOp
    from sarus_data_spec.sarus_statistics.ops.std.op import StdOp
    from sarus_data_spec.sarus_statistics.ops.sum.op import SumOp
    from sarus_data_spec.sarus_statistics.protobuf.multiplicity_pb2 import (
        MultiplicityParameters,
    )
except ModuleNotFoundError as e_pandas_dp:
    if "sarus_data_spec.sarus_statistics" not in str(e_pandas_dp):
        raise

try:
    from sarus_differential_privacy.query import ComposedPrivateQuery
except ModuleNotFoundError as e_pandas_dp:
    if "sarus_differential_privacy" not in str(e_pandas_dp):
        raise

try:
    from sarus_data_spec.sarus_query_builder.builders.bounds_builder import (
        simple_bounds_builder,
    )
    from sarus_data_spec.sarus_query_builder.builders.composed_builder import (
        ComposedBuilder,
        simple_composed_builder,
    )
    from sarus_data_spec.sarus_query_builder.builders.corr_builder import (
        corr_builder as simple_corr_builder,
    )
    from sarus_data_spec.sarus_query_builder.builders.max_multiplicity_builder import (  # noqa : E501
        simple_max_multiplicity_builder,
    )
    from sarus_data_spec.sarus_query_builder.builders.mean_builder import (
        mean_builder,
    )
    from sarus_data_spec.sarus_query_builder.builders.median_builder import (
        median_builder,
    )
    from sarus_data_spec.sarus_query_builder.builders.standard_mechanisms_builder import (  # noqa : E501
        laplace_builder,
    )
    from sarus_data_spec.sarus_query_builder.builders.std_builder import (
        std_builder,
    )
    from sarus_data_spec.sarus_query_builder.builders.sum_builder import (
        sum_builder,
    )
    from sarus_data_spec.sarus_query_builder.core.core import (
        OptimizableQueryBuilder,
    )
    from sarus_data_spec.sarus_query_builder.protobuf.query_pb2 import (
        GenericTask,
        Query,
    )
except ModuleNotFoundError as e_pandas_dp:
    if "sarus" not in str(e_pandas_dp):
        raise
    OptimizableQueryBuilder = t.Any  # type: ignore


NUMERIC_TYPES = ("integer", "float", "boolean")


class pd_shape_dp(ExternalOpImplementation):
    _transform_id = "pandas.PD_SHAPE_DP"
    _non_dp_equivalent_id = "pandas.PD_SHAPE"

    def is_dp(self, signature: SarusBoundSignature) -> bool:
        return True

    async def query_builder(
        self, signature: SarusBoundSignature
    ) -> OptimizableQueryBuilder:
        parent_ds = signature["this"].static_value()
        parent_schema = await parent_ds.manager().async_schema(parent_ds)
        max_mult = await retrieve_max_mult_info_from_parent(
            parent_ds, parent_schema
        )
        max_max_mult = ceil(
            float(
                parent_schema.properties().get(
                    MAX_MAX_MULT, DEFAULT_MAX_MAX_MULT
                )
            )
        )
        single_builder = laplace_builder(
            parent_ds,
            Query(laplace_mechanism=Query.LaplaceMechanism()),
        )
        return builder_from_multiplicity(
            single_builder, max_mult, max_max_mult
        )

    async def private_queries_and_task(
        self, signature: SarusBoundSignature
    ) -> t.Tuple[t.List[st.PrivateQuery], st.Task]:
        """Return the PrivateQueries summarizing DP characteristics."""
        epsilon, delta = await get_budget(signature)
        builder = await self.query_builder(signature)
        tasks = builder.build_query(
            builder.target([(epsilon, delta)], (0, epsilon))
        )
        query = builder.private_query(tasks)
        if isinstance(query, ComposedPrivateQuery):
            return list(query.all_subqueries()), tasks
        else:
            return [query], tasks

    async def call_dp(self, signature: SarusBoundSignature) -> t.Any:
        """Implementation of DP shape.

        A DP implementation receives additional arguments compared to a
        standard external implementation:
            - `budget`: a list of sp.Scalar.PrivacyParams.Point
                object containing each an epsilon and a delta values
            - `seed`: an integer used to parametrize random number generators
            - `pe`: the protected entity used by
            `sarus_data_spec.sarus_statistics`primitives
        """
        # Evaluate arguments
        parent_ds = signature.static_kwargs()["this"]
        (dataframe, budget, seed) = await signature.collect_args()

        # Get QB task parametrization
        _, tasks = await self.private_queries_and_task(signature)

        epsilon = budget[0].epsilon
        shape = dataframe.shape
        parent_schema = await parent_ds.manager().async_schema(parent_ds)
        max_mult = await retrieve_max_mult_info_from_parent(
            parent_ds, parent_schema
        )
        random_generator = np.random.default_rng(abs(seed))
        max_mul, shape_task = retrieve_or_compute_max_mult(
            max_mult, tasks, parent_ds, random_generator, epsilon
        )
        assert len(shape_task) == 1
        n_rows = CountOp(
            parent_ds,
            noise=shape_task[0].parameters["noise"],
        ).value(None, max_mul, random_generator)

        dp_shape = (n_rows, *shape[1:])
        return dp_shape


class pd_sum_dp(ExternalOpImplementation):
    _transform_id = "pandas.PD_SUM_DP"
    _non_dp_equivalent_id = "pandas.PD_SUM"

    def is_dp(self, signature: SarusBoundSignature) -> bool:
        """`axis = 0` and `numeric_only` is `True`"""
        if signature["this"].python_type() == str(pd.Series):
            return True

        axis = signature["axis"].static_value()
        numeric_only = signature["numeric_only"].static_value()
        return (axis == 0 or axis is None) and (numeric_only is True)

    async def query_builder(
        self, signature: SarusBoundSignature
    ) -> OptimizableQueryBuilder:
        parent_ds = signature["this"].static_value()

        parent_schema = await parent_ds.manager().async_schema(parent_ds)
        column_bounds_sum_builder: OptimizableQueryBuilder = (
            simple_composed_builder(
                parent_ds,
                [
                    simple_bounds_builder(
                        parent_ds, Query(bounds=Query.Bounds())
                    ),
                    sum_builder(parent_ds, Query(sum=Query.Sum())),
                ],
            )
        )
        n_cols = len(parent_schema.data_type().children())
        single_builder = simple_composed_builder(
            parent_ds, n_cols * [column_bounds_sum_builder]
        )
        max_mult = await retrieve_max_mult_info_from_parent(
            parent_ds, parent_schema
        )
        max_max_mult = ceil(
            float(
                parent_schema.properties().get(
                    MAX_MAX_MULT, DEFAULT_MAX_MAX_MULT
                )
            )
        )
        return builder_from_multiplicity(
            single_builder, max_mult, max_max_mult
        )

    async def private_queries_and_task(
        self, signature: SarusBoundSignature
    ) -> t.Tuple[t.List[st.PrivateQuery], st.Task]:
        epsilon, delta = await get_budget(signature)
        builder = await self.query_builder(signature=signature)
        tasks = builder.build_query(
            builder.target([(epsilon, delta)], (0, epsilon))
        )
        query = builder.private_query(tasks)
        composed_query = t.cast(ComposedPrivateQuery, query)
        return list(composed_query.all_subqueries()), tasks

    async def call_dp(self, signature: SarusBoundSignature) -> t.Any:
        # Evaluate arguments
        parent_ds: st.Dataset = signature["this"].static_value()
        budget_value = await signature["budget"].collect()
        seed_value = await signature["seed"].collect()
        skipna = await signature["skipna"].collect()

        _, tasks = await self.private_queries_and_task(signature)

        epsilon = budget_value[0].epsilon
        parent_schema = await parent_ds.manager().async_schema(parent_ds)
        max_mult = await retrieve_max_mult_info_from_parent(
            parent_ds, parent_schema
        )
        random_generator = np.random.default_rng(abs(seed_value))
        max_mul, unwrapped_other_tasks = retrieve_or_compute_max_mult(
            max_mult, tasks, parent_ds, random_generator, epsilon
        )
        sum_dp_dict = {}
        schema_fields = parent_schema.data_type().children()
        for tasks_column, (column_name, column_type) in zip(
            unwrapped_other_tasks, (schema_fields.items())
        ):
            bounds_parameters, sum_parameters = (
                t.cast(GenericTask, unwrap(tasks_column.subtasks[0])),
                t.cast(GenericTask, unwrap(tasks_column.subtasks[0])),
            )
            type_pb = column_type.protobuf()
            if type_pb.WhichOneof("type") == "optional":
                if not skipna:
                    sum_dp_dict[column_name] = np.nan
                    continue
                type_pb = type_pb.optional.type
            if type_pb.WhichOneof("type") not in NUMERIC_TYPES:
                continue

            # TODO parent_ds doesn't have a size
            # if parent_ds.size().children()[column_name].size()
            #  < min_count:
            #     sum_dp_dict[column_name] = np.nan
            #     continue

            bounds = BoundOp(
                parent_ds,
                bounds_parameters.parameters["noise"],
            ).value(column_name, max_mul, random_generator)
            sum_op = SumOp(
                parent_ds,
                sum_parameters.parameters["noise"],
            )
            sum_dp_dict[column_name] = sum_op.value(
                column_name, max_mul, bounds, random_generator
            )

        if signature["this"].python_type() == str(pd.DataFrame):
            sum_dp = pd.Series(sum_dp_dict)
        else:
            sum_dp = list(sum_dp_dict.values()).pop()

        return sum_dp


class pd_mean_dp(ExternalOpImplementation):
    _transform_id = "pandas.PD_MEAN_DP"
    _non_dp_equivalent_id = "pandas.PD_MEAN"

    def is_dp(self, signature: SarusBoundSignature) -> bool:
        """`axis = 0` and `numeric_only` is `True`"""
        if signature["this"].python_type() == str(pd.Series):
            return True

        axis = signature["axis"].static_value()
        numeric_only = signature["numeric_only"].static_value()
        return (axis == 0) and (numeric_only is True)

    async def query_builder(
        self, signature: SarusBoundSignature
    ) -> OptimizableQueryBuilder:
        parent_ds = signature["this"].static_value()

        parent_schema = await parent_ds.manager().async_schema(parent_ds)
        max_max_mult = ceil(
            float(
                parent_schema.properties().get(
                    MAX_MAX_MULT, DEFAULT_MAX_MAX_MULT
                )
            )
        )
        max_mult = await retrieve_max_mult_info_from_parent(
            parent_ds, parent_schema
        )
        n_cols = len(parent_schema.data_type().children())
        column_bounds_mean_builder: OptimizableQueryBuilder = (
            simple_composed_builder(
                parent_ds,
                [
                    simple_bounds_builder(
                        parent_ds, Query(bounds=Query.Bounds())
                    ),
                    mean_builder(parent_ds, Query(mean=Query.Mean())),
                ],
            )
        )
        builder = simple_composed_builder(
            parent_ds,
            n_cols * [column_bounds_mean_builder],
        )
        return builder_from_multiplicity(builder, max_mult, max_max_mult)

    async def private_queries_and_task(
        self, signature: SarusBoundSignature
    ) -> t.Tuple[t.List[st.PrivateQuery], st.Task]:
        epsilon, delta = await get_budget(signature)
        builder = await self.query_builder(signature=signature)
        tasks = builder.build_query(
            builder.target([(epsilon, delta)], (0, epsilon))
        )
        query = builder.private_query(tasks)
        composed_query = t.cast(ComposedPrivateQuery, query)
        return list(composed_query.all_subqueries()), tasks

    async def call_dp(self, signature: SarusBoundSignature) -> t.Any:
        # Evaluate arguments
        parent_ds: st.Dataset = signature["this"].static_value()
        budget_value = await signature["budget"].collect()
        seed_value = await signature["seed"].collect()
        skipna = await signature["skipna"].collect()

        _, tasks = await self.private_queries_and_task(signature)

        epsilon = budget_value[0].epsilon
        parent_schema = await parent_ds.manager().async_schema(parent_ds)
        max_mult = await retrieve_max_mult_info_from_parent(
            parent_ds, parent_schema
        )
        random_generator = np.random.default_rng(abs(seed_value))
        max_mul, unwrapped_other_tasks = retrieve_or_compute_max_mult(
            max_mult, tasks, parent_ds, random_generator, epsilon
        )
        mean_dp_dict = {}
        schema_fields = parent_schema.data_type().children()
        for tasks_column, (column_name, column_type) in zip(
            unwrapped_other_tasks, (schema_fields.items())
        ):
            bounds_parameters, mean_parameters = (
                t.cast(GenericTask, unwrap(tasks_column.subtasks[0])),
                t.cast(GenericTask, unwrap(tasks_column.subtasks[1])),
            )

            type_pb = column_type.protobuf()
            if type_pb.WhichOneof("type") == "optional":
                if not skipna:
                    mean_dp_dict[column_name] = np.nan
                    continue
                type_pb = type_pb.optional.type
            if type_pb.WhichOneof("type") not in NUMERIC_TYPES:
                continue
            bounds = BoundOp(
                parent_ds,
                bounds_parameters.parameters["noise"],
            ).value(column_name, max_mul, random_generator)
            mean_op = MeanOp(
                parent_ds,
                mean_parameters.parameters["noise"],
            )
            mean_dp_dict[column_name] = mean_op.value(
                column_name, max_mul, bounds, random_generator
            )

        if signature["this"].python_type() == str(pd.DataFrame):
            mean_dp = pd.Series(mean_dp_dict)
        else:
            mean_dp = list(mean_dp_dict.values()).pop()

        return mean_dp


class pd_median_dp(ExternalOpImplementation):
    _transform_id = "pandas.PD_MEDIAN_DP"
    _non_dp_equivalent_id = "pandas.PD_MEDIAN"

    def is_dp(self, signature: SarusBoundSignature) -> bool:
        """`axis = 0` and `numeric_only` is `True`"""
        if signature["this"].python_type() == str(pd.Series):
            return True

        axis = signature["axis"].static_value()
        numeric_only = signature["numeric_only"].static_value()
        return axis == 0 and numeric_only is True

    async def query_builder(
        self, signature: SarusBoundSignature
    ) -> OptimizableQueryBuilder:
        parent_ds = signature["this"].static_value()
        parent_schema = await parent_ds.manager().async_schema(parent_ds)
        max_max_mult = ceil(
            float(
                parent_schema.properties().get(
                    MAX_MAX_MULT, DEFAULT_MAX_MAX_MULT
                )
            )
        )
        max_mult = await retrieve_max_mult_info_from_parent(
            parent_ds, parent_schema
        )
        n_cols = len(parent_schema.data_type().children())
        column_bounds_median_builder: OptimizableQueryBuilder = (
            simple_composed_builder(
                parent_ds,
                [
                    simple_bounds_builder(
                        parent_ds, Query(bounds=Query.Bounds())
                    ),
                    median_builder(parent_ds, Query(median=Query.Median())),
                ],
            )
        )
        builder = simple_composed_builder(
            parent_ds,
            n_cols * [column_bounds_median_builder],
        )
        return builder_from_multiplicity(builder, max_mult, max_max_mult)

    async def private_queries_and_task(
        self, signature: SarusBoundSignature
    ) -> t.Tuple[t.List[st.PrivateQuery], st.Task]:
        epsilon, delta = await get_budget(signature)
        builder = await self.query_builder(signature=signature)
        tasks = builder.build_query(
            builder.target([(epsilon, delta)], (0, epsilon))
        )
        query = builder.private_query(tasks)
        composed_query = t.cast(ComposedPrivateQuery, query)
        return list(composed_query.all_subqueries()), tasks

    async def call_dp(self, signature: SarusBoundSignature) -> t.Any:
        # Evaluate arguments
        parent_ds: st.Dataset = signature["this"].static_value()
        budget_value = await signature["budget"].collect()
        seed_value = await signature["seed"].collect()
        skipna = await signature["skipna"].collect()

        _, tasks = await self.private_queries_and_task(signature)
        epsilon = budget_value[0].epsilon

        parent_schema = await parent_ds.manager().async_schema(parent_ds)
        max_mult = await retrieve_max_mult_info_from_parent(
            parent_ds, parent_schema
        )
        random_generator = np.random.default_rng(abs(seed_value))
        max_mul, unwrapped_other_tasks = retrieve_or_compute_max_mult(
            max_mult, tasks, parent_ds, random_generator, epsilon
        )

        median_dp_dict = {}
        schema_fields = parent_schema.data_type().children()
        for tasks_column, (column_name, column_type) in zip(
            unwrapped_other_tasks, (schema_fields.items())
        ):
            bounds_parameters, median_parameters = (
                t.cast(GenericTask, unwrap(tasks_column.subtasks[0])),
                t.cast(GenericTask, unwrap(tasks_column.subtasks[1])),
            )
            type_pb = column_type.protobuf()
            if type_pb.WhichOneof("type") == "optional":
                if not skipna:
                    median_dp_dict[column_name] = np.nan
                    continue
                type_pb = type_pb.optional.type

            if type_pb.WhichOneof("type") not in NUMERIC_TYPES:
                continue

            bounds = BoundOp(
                parent_ds,
                bounds_parameters.parameters["noise"],
            ).value(column_name, max_mul, random_generator)
            median_op = MedianOp(
                parent_ds,
                median_parameters.parameters["noise"],
            )
            median_dp_dict[column_name] = median_op.value(
                column_name, max_mul, bounds, random_generator
            )
        if signature["this"].python_type() == str(pd.DataFrame):
            median_dp = pd.Series(median_dp_dict)
        else:
            median_dp = list(median_dp_dict.values()).pop()

        return median_dp


class pd_std_dp(ExternalOpImplementation):
    _transform_id = "pandas.PD_STD_DP"
    _non_dp_equivalent_id = "pandas.PD_STD"

    def is_dp(self, signature: SarusBoundSignature) -> bool:
        """`axis = 0` and `numeric_only` is `True`"""
        if signature["this"].python_type() == str(pd.Series):
            return True

        axis = signature["axis"].static_value()
        numeric_only = signature["numeric_only"].static_value()
        return axis == 0 and numeric_only is True

    async def query_builder(
        self, signature: SarusBoundSignature
    ) -> OptimizableQueryBuilder:
        parent_ds = signature["this"].static_value()

        parent_schema = await parent_ds.manager().async_schema(parent_ds)
        max_max_mult = ceil(
            float(
                parent_schema.properties().get(
                    MAX_MAX_MULT, DEFAULT_MAX_MAX_MULT
                )
            )
        )
        max_mult = await retrieve_max_mult_info_from_parent(
            parent_ds, parent_schema
        )
        n_cols = len(parent_schema.data_type().children())
        column_bounds_std_builder: OptimizableQueryBuilder = (
            simple_composed_builder(
                parent_ds,
                [
                    simple_bounds_builder(
                        parent_ds, Query(bounds=Query.Bounds())
                    ),
                    std_builder(parent_ds, Query(std=Query.Std())),
                ],
            )
        )

        builder = simple_composed_builder(
            parent_ds,
            n_cols * [column_bounds_std_builder],
        )
        return builder_from_multiplicity(builder, max_mult, max_max_mult)

    async def private_queries_and_task(
        self, signature: SarusBoundSignature
    ) -> t.Tuple[t.List[st.PrivateQuery], st.Task]:
        epsilon, delta = await get_budget(signature)
        builder = await self.query_builder(signature=signature)
        tasks = builder.build_query(
            builder.target([(epsilon, delta)], (0, epsilon))
        )
        query = builder.private_query(tasks)
        composed_query = t.cast(ComposedPrivateQuery, query)
        return list(composed_query.all_subqueries()), tasks

    async def call_dp(self, signature: SarusBoundSignature) -> t.Any:
        # Evaluate arguments
        parent_ds: st.Dataset = signature["this"].static_value()
        budget_value = await signature["budget"].collect()
        seed_value = await signature["seed"].collect()
        skipna = await signature["skipna"].collect()

        _, tasks = await self.private_queries_and_task(signature)
        epsilon = budget_value[0].epsilon
        parent_schema = await parent_ds.manager().async_schema(parent_ds)
        max_mult = await retrieve_max_mult_info_from_parent(
            parent_ds, parent_schema
        )
        random_generator = np.random.default_rng(abs(seed_value))
        max_mul, unwrapped_other_tasks = retrieve_or_compute_max_mult(
            max_mult, tasks, parent_ds, random_generator, epsilon
        )

        std_dp_dict = {}
        schema_fields = parent_schema.data_type().children()
        for tasks_column, (column_name, column_type) in zip(
            unwrapped_other_tasks, (schema_fields.items())
        ):
            bounds_parameters, std_parameters = (
                t.cast(GenericTask, unwrap(tasks_column.subtasks[0])),
                t.cast(GenericTask, unwrap(tasks_column.subtasks[0])),
            )
            type_pb = column_type.protobuf()
            if type_pb.HasField("optional"):
                if not skipna:
                    std_dp_dict[column_name] = np.nan
                    continue
                type_pb = type_pb.optional.type
            if type_pb.WhichOneof("type") not in NUMERIC_TYPES:
                continue

            bounds = BoundOp(
                parent_ds,
                bounds_parameters.parameters["noise"],
            ).value(column_name, max_mul, random_generator)
            std_op = StdOp(
                parent_ds,
                std_parameters.parameters["noise_mean"],
                std_parameters.parameters["noise_square"],
                std_parameters.parameters["noise_count"],
            )
            std_dp_dict[column_name] = std_op.value(
                column_name, max_mul, bounds, random_generator
            )

        if signature["this"].python_type() == str(pd.DataFrame):
            std_dp = pd.Series(std_dp_dict)
        else:
            std_dp = list(std_dp_dict.values()).pop()

        return std_dp


class pd_count_dp(ExternalOpImplementation):
    _transform_id = "pandas.PD_COUNT_DP"
    _non_dp_equivalent_id = "pandas.PD_COUNT"

    def is_dp(self, signature: SarusBoundSignature) -> bool:
        """`axis = 0` and `numeric_only` is `True`"""
        axis = signature["axis"].static_value()
        return bool(axis == 0)

    async def query_builder(
        self, signature: SarusBoundSignature
    ) -> OptimizableQueryBuilder:
        parent_ds = signature["this"].static_value()

        parent_schema = await parent_ds.manager().async_schema(parent_ds)

        max_max_mult = ceil(
            float(
                parent_schema.properties().get(
                    MAX_MAX_MULT, DEFAULT_MAX_MAX_MULT
                )
            )
        )
        max_mult = await retrieve_max_mult_info_from_parent(
            parent_ds, parent_schema
        )
        n_cols = len(parent_schema.data_type().children())
        column_counts_builder = laplace_builder(
            parent_ds, Query(laplace_mechanism=Query.LaplaceMechanism())
        )
        builder = simple_composed_builder(
            parent_ds, n_cols * [column_counts_builder]
        )
        return builder_from_multiplicity(builder, max_mult, max_max_mult)

    async def private_queries_and_task(
        self, signature: SarusBoundSignature
    ) -> t.Tuple[t.List[st.PrivateQuery], st.Task]:
        epsilon, delta = await get_budget(signature)
        builder = await self.query_builder(signature=signature)
        tasks = builder.build_query(
            builder.target([(epsilon, delta)], (0, epsilon))
        )
        query = builder.private_query(tasks)
        composed_query = t.cast(ComposedPrivateQuery, query)
        return list(composed_query.all_subqueries()), tasks

    async def call_dp(self, signature: SarusBoundSignature) -> t.Any:
        # Evaluate arguments
        parent_ds: st.Dataset = signature["this"].static_value()
        budget_value = await signature["budget"].collect()
        seed_value = await signature["seed"].collect()

        _, tasks = await self.private_queries_and_task(signature)

        epsilon = budget_value[0].epsilon
        parent_schema = await parent_ds.manager().async_schema(parent_ds)
        max_mult = await retrieve_max_mult_info_from_parent(
            parent_ds, parent_schema
        )
        random_generator = np.random.default_rng(abs(seed_value))
        max_mul, unwrapped_other_tasks = retrieve_or_compute_max_mult(
            max_mult, tasks, parent_ds, random_generator, epsilon
        )
        count_dp_dict = {}

        schema_fields = parent_schema.data_type().children()
        for count_parameter, column_name in zip(
            unwrapped_other_tasks, (schema_fields.keys())
        ):
            count_dp_dict[column_name] = CountOp(
                parent_ds,
                count_parameter.parameters["noise"],
            ).value(column_name, max_mul, random_generator)

        if signature["this"].python_type() == str(pd.DataFrame):
            count_dp = pd.Series(count_dp_dict)
        else:
            count_dp = list(count_dp_dict.values()).pop()
        return count_dp


class pd_value_counts_dp(ExternalOpImplementation):
    _transform_id: str = "pandas.PD_VALUE_COUNTS_DP"
    _non_dp_equivalent_id = "pandas.PD_VALUE_COUNTS"

    def is_dp(self, signature: SarusBoundSignature) -> bool:
        return signature["this"].python_type() == str(pd.Series)

    async def query_builder(
        self, signature: SarusBoundSignature
    ) -> OptimizableQueryBuilder:
        parent_ds = signature["this"].static_value()

        parent_schema = await parent_ds.manager().async_schema(parent_ds)
        max_max_mult = ceil(
            float(
                parent_schema.properties().get(
                    MAX_MAX_MULT, DEFAULT_MAX_MAX_MULT
                )
            )
        )
        max_mult = await retrieve_max_mult_info_from_parent(
            parent_ds, parent_schema
        )
        n_cols = len(parent_schema.data_type().children())
        column_counts_builder = laplace_builder(
            parent_ds, Query(laplace_mechanism=Query.LaplaceMechanism())
        )
        builder = simple_composed_builder(
            parent_ds, n_cols * [column_counts_builder]
        )
        return builder_from_multiplicity(builder, max_mult, max_max_mult)

    async def private_queries_and_task(
        self, signature: SarusBoundSignature
    ) -> t.Tuple[t.List[st.PrivateQuery], st.Task]:
        """Return the PrivateQueries summarizing DP characteristics."""
        epsilon, delta = await get_budget(signature)
        builder = await self.query_builder(signature=signature)
        tasks = builder.build_query(
            builder.target([(epsilon, delta)], (0, epsilon))
        )
        query = builder.private_query(tasks)
        composed_query = t.cast(ComposedPrivateQuery, query)
        return list(composed_query.all_subqueries()), tasks

    async def call_dp(self, signature: SarusBoundSignature) -> t.Any:
        """Implementation of DP shape.

        A DP implementation receives additional arguments compared to a
        standard external implementation:
            - `budget`: a list of sp.Scalar.PrivacyParams.Point
                object containing each an epsilon and a delta values
            - `seed`: an integer used to parametrize random number generators
            - `pe`: the protected entity used by
            `sarus_data_spec.sarus_statistics` primitives
        """
        parent_ds: st.Dataset = signature["this"].static_value()
        budget_value = await signature["budget"].collect()
        seed_value = await signature["seed"].collect()
        sort = await signature["sort"].collect()
        ascending = await signature["ascending"].collect()
        normalize = await signature["normalize"].collect()
        dropna = await signature["dropna"].collect()

        _, tasks = await self.private_queries_and_task(signature)

        epsilon = budget_value[0].epsilon
        random_generator = np.random.default_rng(abs(seed_value))
        parent_schema = parent_ds.schema()
        max_mult = await retrieve_max_mult_info_from_parent(
            parent_ds, parent_schema
        )
        max_mul, unwrapped_other_tasks = retrieve_or_compute_max_mult(
            max_mult, tasks, parent_ds, random_generator, epsilon
        )

        count_dp_dict = {}
        schema_fields = parent_schema.data_type().children()
        for count_parameters, column_name in zip(
            unwrapped_other_tasks, (schema_fields.keys())
        ):
            histogram = t.cast(
                t.Dict[str, float],
                HistogramOp(
                    parent_ds,
                    noise=count_parameters.parameters["noise"],
                    sort=sort,
                    ascending=ascending,
                    normalize=normalize,
                    dropna=dropna,
                ).value(column_name, max_mul, random_generator),
            )
            count_dp_dict[column_name] = histogram

        # for now only on series
        values = list(count_dp_dict.values())[0]
        name = list(count_dp_dict.keys())[0]
        # added to be conform to its expected schema
        count_dp = pd.Series(data=values, name=name)
        return count_dp


class pd_corr_dp(ExternalOpImplementation):
    _transform_id = "pandas.PD_CORR_DP"
    _non_dp_equivalent_id = "pandas.PD_CORR"

    def is_dp(self, signature: SarusBoundSignature) -> bool:
        """
        `numeric_only` is `True`
        DataFrame
        `method = Pearson`
        `min_periods = 1`
        """
        method = signature["method"].static_value()
        min_periods = signature["min_periods"].static_value()
        is_dataframe = bool(
            signature["this"].python_type() == str(pd.DataFrame)
        )
        return is_dataframe and (min_periods == 1) and (method == "pearson")

    async def query_builder(
        self, signature: SarusBoundSignature
    ) -> OptimizableQueryBuilder:
        parent_ds = signature["this"].static_value()
        parent_schema = await parent_ds.manager().async_schema(parent_ds)
        max_max_mult = ceil(
            float(
                parent_schema.properties().get(
                    MAX_MAX_MULT, DEFAULT_MAX_MAX_MULT
                )
            )
        )
        max_mult = await retrieve_max_mult_info_from_parent(
            parent_ds, parent_schema
        )
        corr_builder: OptimizableQueryBuilder = simple_corr_builder(
            parent_ds, Query(covariance=Query.Covariance())
        )

        return builder_from_multiplicity(corr_builder, max_mult, max_max_mult)

    async def private_queries_and_task(
        self, signature: SarusBoundSignature
    ) -> t.Tuple[t.List[st.PrivateQuery], st.Task]:
        epsilon, delta = await get_budget(signature)
        builder = await self.query_builder(signature=signature)
        tasks = builder.build_query(
            builder.target([(epsilon, delta)], (0, epsilon))
        )
        query = t.cast(ComposedPrivateQuery, builder.private_query(tasks))
        if isinstance(query, ComposedPrivateQuery):
            return list(query.all_subqueries()), tasks
        else:
            return [query], tasks

    async def call_dp(self, signature: SarusBoundSignature) -> t.Any:
        # Evaluate arguments
        parent_ds: st.Dataset = signature["this"].static_value()
        budget_value = await signature["budget"].collect()
        seed_value = await signature["seed"].collect()

        _, tasks = await self.private_queries_and_task(signature)
        epsilon = budget_value[0].epsilon
        # Compute DP value
        random_generator = np.random.default_rng(abs(seed_value))
        parent_schema = parent_ds.schema()
        max_mult = await retrieve_max_mult_info_from_parent(
            parent_ds, parent_schema
        )
        max_mul, unwrapped_other_tasks = retrieve_or_compute_max_mult(
            max_mult, tasks, parent_ds, random_generator, epsilon
        )
        assert len(unwrapped_other_tasks) == 1
        corr_dp = CorrOp(
            parent_ds,
            unwrapped_other_tasks[0].parameters["epsilon"],
            unwrapped_other_tasks[0].parameters["dims"],
        ).value(max_mul, random_generator=random_generator)

        return corr_dp


async def get_budget(signature: SarusBoundSignature) -> t.Tuple[float, float]:
    """Retrieve (epsilon, delta) from signature"""
    budget = await signature["budget"].collect()
    if len(budget) != 1:
        raise NotImplementedError(
            "The PrivacyParams contains more than 1 point in the privacy "
            "profile."
        )

    epsilon = budget[0].epsilon
    delta = budget[0].delta
    if epsilon == 0.0:
        raise ValueError("`epsilon` should be greater than 0.")

    return epsilon, delta


def builder_from_multiplicity(
    builder: OptimizableQueryBuilder, max_mult: str, max_max_mult: int
) -> OptimizableQueryBuilder:
    """Adds a builder from multiplicity if the max_mult is empty"""

    if max_mult != "":
        # in this case no need to recompute it
        return builder
    max_mult_builder = simple_max_multiplicity_builder(
        builder.dataset,
        Query(
            max_multiplicity=Query.MaxMultiplicity(
                max_max_multiplicity=max_max_mult
            )
        ),
    )
    # need to recompute it, take MAX_MAX_MULT as bound
    if isinstance(builder, ComposedBuilder):
        builder._builders.insert(0, max_mult_builder)  # type:ignore
        builder._weights.insert(0, 1)  # type:ignore
    else:
        builder = simple_composed_builder(
            builder.dataset,
            [max_mult_builder, builder],
        )
    return builder


def retrieve_or_compute_max_mult(
    max_mult: str,
    tasks: st.Task,
    parent_ds: st.Dataset,
    random_generator: np.random.Generator,
    epsilon: float,
) -> t.Tuple[int, t.List[st.Task]]:
    """Either retrieves the max_mult if it is not an empty string or computes it
    using the tasks"""

    if max_mult == "":
        # there is no pre-computed max_multiplicity so we need
        # to recompute it
        unwrapped = [unwrap(subtask) for subtask in tasks.subtasks]
        max_mult_task, unwrapped_other_tasks = (
            t.cast(MultiplicityParameters, unwrapped[0]),
            t.cast(t.List[GenericTask], unwrapped[1:]),
        )
        max_mul = MaxMultiplicityOp(
            parent_ds,
            epsilon,  # parameter for quantiles
            max_mult_task.compute.noise_user_count,
            max_mult_task.compute.noise_multiplicity,
            max_mult_task.compute.max_max_multiplicity,
        ).value(random_generator)
    else:
        max_mul = float(max_mult)
        if hasattr(tasks, "subtasks"):
            unwrapped_other_tasks = t.cast(
                t.List[GenericTask],
                [unwrap(subtask) for subtask in tasks.subtasks],
            )  # for type compatibility mainly
        else:
            unwrapped_other_tasks = [tasks]
    return ceil(max_mul), unwrapped_other_tasks



from typing import Any, Dict, Iterable, Optional, Sequence, Union

import numpy as np
import pandas as pd

from sarus_data_spec.dataspec_validator.parameter_kind import DATASPEC, STATIC
from sarus_data_spec.dataspec_validator.signature import (
    SarusParameter,
    SarusSignature,
    SarusSignatureValue,
)

from .external_op import ExternalOpImplementation

try:
    from optbinning import BinningProcess, OptimalBinning, Scorecard
    from optbinning.binning.binning_statistics import BinningTable
except ModuleNotFoundError:
    OptimalBinning = Any
    Scorecard = Any
    BinningTable = Any
    BinningProcess = Any


class optimal_binning(ExternalOpImplementation):
    _transform_id = "optbinning.OPTIMAL_BINNING"
    _signature = SarusSignature(
        SarusParameter(name="name", annotation=Optional[str], default=""),
        SarusParameter(
            name="dtype", annotation=Optional[str], default="numerical"
        ),
        SarusParameter(
            name="prebinning_method", annotation=Optional[str], default="cart"
        ),
        SarusParameter(name="solver", annotation=Optional[str], default="cp"),
        SarusParameter(
            name="divergence", annotation=Optional[str], default="iv"
        ),
        SarusParameter(
            name="max_n_prebins", annotation=Optional[int], default=20
        ),
        SarusParameter(
            name="min_prebin_size", annotation=Optional[float], default=0.05
        ),
        SarusParameter(
            name="min_n_bins",
            annotation=Optional[Union[int, None]],
            default=None,
        ),
        SarusParameter(
            name="max_n_bins",
            annotation=Optional[Union[int, None]],
            default=None,
        ),
        SarusParameter(
            name="min_bin_size",
            annotation=Optional[Union[float, None]],
            default=None,
        ),
        SarusParameter(
            name="max_bin_size",
            annotation=Optional[Union[float, None]],
            default=None,
        ),
        SarusParameter(
            name="min_bin_n_nonevent",
            annotation=Optional[Union[int, None]],
            default=None,
        ),
        SarusParameter(
            name="max_bin_n_nonevent",
            annotation=Optional[Union[int, None]],
            default=None,
        ),
        SarusParameter(
            name="min_bin_n_event",
            annotation=Optional[Union[int, None]],
            default=None,
        ),
        SarusParameter(
            name="max_bin_n_event",
            annotation=Optional[Union[int, None]],
            default=None,
        ),
        SarusParameter(
            name="monotonic_trend", annotation=Optional[str], default="auto"
        ),
        SarusParameter(
            name="min_event_rate_diff", annotation=Optional[float], default=0
        ),
        SarusParameter(
            name="max_pvalue",
            annotation=Optional[Union[float, None]],
            default=None,
        ),
        SarusParameter(
            name="max_pvalue_policy",
            annotation=Optional[str],
            default="consecutive",
        ),
        SarusParameter(name="gamma", annotation=Optional[float], default=0),
        SarusParameter(
            name="outlier_detector",
            annotation=Optional[Union[str, None]],
            default=None,
        ),
        SarusParameter(
            name="outlier_params",
            annotation=Optional[Union[Dict, None]],
            default=None,
        ),
        SarusParameter(
            name="class_weight",
            annotation=Optional[Union[Dict, str, None]],
            default=None,
        ),
        SarusParameter(
            name="cat_cutoff",
            annotation=Optional[Union[float, None]],
            default=None,
        ),
        SarusParameter(
            name="cat_unknown",
            annotation=Optional[Union[float, str, None]],
            default=None,
        ),
        SarusParameter(
            name="user_splits",
            annotation=Optional[Union[Iterable, None]],
            default=None,
        ),
        SarusParameter(
            name="user_splits_fixed",
            annotation=Optional[Union[Iterable, None]],
            default=None,
        ),
        SarusParameter(
            name="special_codes",
            annotation=Optional[Union[Iterable, Dict, None]],
            default=None,
        ),
        SarusParameter(
            name="split_digits",
            annotation=Optional[Union[int, None]],
            default=None,
        ),
        SarusParameter(
            name="mip_solver", annotation=Optional[str], default="bop"
        ),
        SarusParameter(
            name="time_limit", annotation=Optional[int], default=100
        ),
        SarusParameter(
            name="verbose", annotation=Optional[bool], default=False
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return OptimalBinning(**kwargs)


class optimal_binning_fit(ExternalOpImplementation):
    _transform_id = "optbinning.OPTIMAL_BINNING_FIT"
    _signature = SarusSignature(
        SarusParameter(
            name="opt_binning",
            annotation=OptimalBinning,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="x", annotation=np.ndarray, condition=DATASPEC | STATIC
        ),
        SarusParameter(
            name="y", annotation=np.ndarray, condition=DATASPEC | STATIC
        ),
        SarusParameter(
            name="sample_weight", annotation=Optional[np.ndarray], default=None
        ),
        SarusParameter(name="check_input", annotation=bool, default=False),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.fit(**kwargs)


class optimal_binning_fit_transform(ExternalOpImplementation):
    _transform_id = "optbinning.OPTIMAL_BINNING_FIT_TRANSFORM"
    _signature = SarusSignature(
        SarusParameter(
            name="opt_binning",
            annotation=OptimalBinning,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="x", annotation=np.ndarray, condition=DATASPEC | STATIC
        ),
        SarusParameter(
            name="y", annotation=np.ndarray, condition=DATASPEC | STATIC
        ),
        SarusParameter(
            name="sample_weight", annotation=Optional[np.ndarray], default=None
        ),
        SarusParameter(name="metric", annotation=str, default="woe"),
        SarusParameter(
            name="metric_special", annotation=Union[float, str], default=0
        ),
        SarusParameter(
            name="metric_missing", annotation=Union[float, str], default=0
        ),
        SarusParameter(
            name="show_digits", annotation=Optional[int], default=2
        ),
        SarusParameter(name="check_input", annotation=bool, default=False),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.fit_transform(**kwargs)


class optimal_binning_transform(ExternalOpImplementation):
    _transform_id = "optbinning.OPTIMAL_BINNING_TRANSFORM"
    _signature = SarusSignature(
        SarusParameter(
            name="opt_binning",
            annotation=OptimalBinning,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(name="x", annotation=np.ndarray),
        SarusParameter(name="metric", annotation=str, default="woe"),
        SarusParameter(
            name="metric_special", annotation=Union[float, str], default=0
        ),
        SarusParameter(
            name="metric_missing", annotation=Union[float, str], default=0
        ),
        SarusParameter(
            name="show_digits", annotation=Optional[int], default=2
        ),
        SarusParameter(name="check_input", annotation=bool, default=False),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.transform(**kwargs)


class optimal_binning_binning_table(ExternalOpImplementation):
    _transform_id = "optbinning.OPTIMAL_BINNING_BINNING_TABLE"
    _signature = SarusSignature(
        SarusParameter(
            name="opt_binning",
            annotation=OptimalBinning,
            condition=DATASPEC | STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        # Extract the OptimalBinning instance
        opt_binning = signature["opt_binning"].value
        bintable = opt_binning.binning_table
        bintable.build()
        # Retrieve the binning table already built
        return bintable


class optbinning_binningProcess(ExternalOpImplementation):
    _transform_id = "optbinning.BINNING_PROCESS"
    _signature = SarusSignature(
        SarusParameter(
            name="variable_names",
            annotation=Sequence[str],
        ),
        SarusParameter(
            name="max_n_prebins",
            annotation=int,
            default=20,
        ),
        SarusParameter(
            name="min_prebin_size",
            annotation=float,
            default=0.05,
        ),
        SarusParameter(
            name="min_n_bins",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="max_n_bins",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="min_bin_size",
            annotation=Optional[float],
            default=None,
        ),
        SarusParameter(
            name="max_bin_size",
            annotation=Optional[float],
            default=None,
        ),
        SarusParameter(
            name="max_pvalue",
            annotation=Optional[float],
            default=None,
        ),
        SarusParameter(
            name="max_pvalue_policy",
            annotation=str,
            default="consecutive",
        ),
        SarusParameter(
            name="selection_criteria",
            annotation=Optional[Dict],
            default=None,
        ),
        SarusParameter(
            name="fixed_variables",
            annotation=Optional[Union[Sequence[str], None]],
            default=None,
        ),
        SarusParameter(
            name="categorical_variables",
            annotation=Optional[Union[Sequence[str], None]],
            default=None,
        ),
        SarusParameter(
            name="special_codes",
            annotation=Optional[Union[Sequence, None]],
            default=None,
        ),
        SarusParameter(
            name="split_digits",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="binning_fit_params",
            annotation=Optional[Dict],
            default=None,
        ),
        SarusParameter(
            name="binning_transform_params",
            annotation=Optional[Dict],
            default=None,
        ),
        SarusParameter(
            name="n_jobs",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="verbose",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return BinningProcess(**kwargs)


class fit(ExternalOpImplementation):
    _transform_id = "optbinning.BINNING_PROCESS_FIT"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=BinningProcess,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="X",
            annotation=Union[np.ndarray, pd.DataFrame],
        ),
        SarusParameter(
            name="y",
            annotation=np.ndarray,
        ),
        SarusParameter(
            name="sample_weight",
            annotation=Optional[np.ndarray],
            default=None,
        ),
        SarusParameter(
            name="check_input",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.fit(**kwargs)


class fit_transform(ExternalOpImplementation):
    _transform_id = "optbinning.BINNING_PROCESS_FIT_TRANSFORM"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=BinningProcess,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="X",
            annotation=Union[np.ndarray, pd.DataFrame],
        ),
        SarusParameter(
            name="y",
            annotation=np.ndarray,
        ),
        SarusParameter(
            name="sample_weight",
            annotation=Optional[np.ndarray],
            default=None,
        ),
        SarusParameter(
            name="metric",
            annotation=Optional[str],
            default=None,
        ),
        SarusParameter(
            name="metric_special",
            annotation=float,
            default=0,
        ),
        SarusParameter(
            name="metric_missing",
            annotation=float,
            default=0,
        ),
        SarusParameter(
            name="show_digits",
            annotation=int,
            default=2,
        ),
        SarusParameter(
            name="check_input",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.fit_transform(**kwargs)


class transform(ExternalOpImplementation):
    _transform_id = "optbinning.BINNING_PROCESS_TRANSFORM"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=BinningProcess,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="X",
            annotation=Union[np.ndarray, pd.DataFrame],
        ),
        SarusParameter(
            name="metric",
            annotation=Optional[str],
            default=None,
        ),
        SarusParameter(
            name="metric_special",
            annotation=float,
            default=0,
        ),
        SarusParameter(
            name="metric_missing",
            annotation=float,
            default=0,
        ),
        SarusParameter(
            name="show_digits",
            annotation=int,
            default=2,
        ),
        SarusParameter(
            name="check_input",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.transform(**kwargs)


class optbinning_binning_table(ExternalOpImplementation):
    _transform_id = "optbinning.BINNING_TABLE"
    _signature = SarusSignature(
        SarusParameter(
            name="name",
            annotation=Optional[str],
            default="",
        ),
        SarusParameter(
            name="dtype",
            annotation=Optional[str],
            default="numerical",
        ),
        SarusParameter(
            name="special_codes",
            annotation=Optional[Union[np.ndarray, dict, None]],
            default=None,
        ),
        SarusParameter(
            name="splits",
            annotation=np.ndarray,
        ),
        SarusParameter(
            name="n_nonevent",
            annotation=np.ndarray,
            condition=STATIC,
        ),
        SarusParameter(
            name="n_event",
            annotation=np.ndarray,
        ),
        SarusParameter(
            name="min_x",
            annotation=Optional[float],
            default=None,
        ),
        SarusParameter(
            name="max_x",
            annotation=Optional[float],
            default=None,
        ),
        SarusParameter(
            name="categories",
            annotation=Optional[Union[list, np.ndarray]],
            default=None,
        ),
        SarusParameter(
            name="cat_others",
            annotation=Optional[Union[list, np.ndarray]],
            default=None,
        ),
        SarusParameter(
            name="user_splits",
            annotation=Optional[np.ndarray],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        # Extract arguments using signature.collect_kwargs()
        kwargs = signature.collect_kwargs()
        return BinningTable(**kwargs)


class optbinning_table_build(ExternalOpImplementation):
    _transform_id = "optbinning.BINNING_TABLE_BUILD"
    _signature = SarusSignature(
        SarusParameter(
            name="binning_table",
            annotation=BinningTable,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="show_digits",
            annotation=Optional[int],
            default=2,
        ),
        SarusParameter(
            name="add_totals",
            annotation=Optional[bool],
            default=True,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this, kwargs) = signature.collect_kwargs_method()
        df = this.build(**kwargs)
        # small edit of the result to convert it in arrow
        if kwargs["add_totals"]:
            df["WoE"]["Totals"] = 0.0
            df["WoE"] = df["WoE"].astype(float)
            df_reset = df.reset_index()
            df_reset["index"] = df_reset["index"].astype(
                str
            )  # Replace str with the desired type
            df_reset = df_reset.set_index("index")
        return df_reset


class scorecard(ExternalOpImplementation):
    _transform_id = "optbinning.SCORECARD"
    _signature = SarusSignature(
        SarusParameter(
            name="binning_process",
            annotation=object,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="estimator",
            annotation=object,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="scaling_method",
            annotation=Optional[str],
            default=None,
        ),
        SarusParameter(
            name="scaling_method_params",
            annotation=Optional[dict],
            default=None,
        ),
        SarusParameter(
            name="intercept_based",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="reverse_scorecard",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="rounding",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="verbose",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return Scorecard(**kwargs)


class scorecard_predict(ExternalOpImplementation):
    _transform_id = "optbinning.SCORECARD_PREDICT"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=scorecard,
            condition=DATASPEC,
        ),
        SarusParameter(
            name="X",
            annotation=pd.DataFrame,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this, kwargs) = signature.collect_kwargs_method()
        return this.predict(**kwargs)


class scorecard_predict_proba(ExternalOpImplementation):
    _transform_id = "optbinning.SCORECARD_PREDICT_PROBA"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=scorecard,
            condition=DATASPEC,
        ),
        SarusParameter(
            name="X",
            annotation=pd.DataFrame,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this, kwargs) = signature.collect_kwargs_method()
        return this.predict_proba(**kwargs)


class scorecard_fit(ExternalOpImplementation):
    _transform_id = "optbinning.SCORECARD_FIT"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=scorecard,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="X",
            annotation=pd.DataFrame,
        ),
        SarusParameter(
            name="y",
            annotation=np.ndarray,
        ),
        SarusParameter(
            name="sample_weight",
            annotation=Optional[np.ndarray],
            default=None,
        ),
        SarusParameter(
            name="metric_special",
            annotation=Union[float, str],
            default=0,
        ),
        SarusParameter(
            name="metric_missing",
            annotation=Union[float, str],
            default=0,
        ),
        SarusParameter(
            name="show_digits",
            annotation=int,
            default=2,
        ),
        SarusParameter(
            name="check_input",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this, kwargs) = signature.collect_kwargs_method()
        return this.fit(**kwargs)



from __future__ import annotations

import typing as t

from sarus_data_spec.base import Referring
from sarus_data_spec.dataset import Dataset
from sarus_data_spec.statistics import Statistics
import sarus_data_spec.protobuf as sp
import sarus_data_spec.typing as st


class Marginals(Referring[sp.Marginals]):
    """A python class to describe marginals"""

    def __init__(self, protobuf: sp.Marginals, store: bool = True) -> None:
        self._referred = {
            protobuf.dataset
        }  # This has to be defined before it is initialized
        super().__init__(protobuf, store=store)

    def prototype(self) -> t.Type[sp.Marginals]:
        """Return the type of the underlying protobuf."""
        return sp.Marginals

    def dataset(self) -> Dataset:
        return t.cast(
            Dataset, self.storage().referrable(self._protobuf.dataset)
        )

    def statistics(self) -> Statistics:
        """returns the python object writing the statistics proto"""
        return Statistics(self.protobuf().statistics)


# Builder
def marginals(
    dataset: st.Dataset,
    statistics: t.Optional[st.Statistics] = None,
    properties: t.Optional[t.Mapping[str, str]] = None,
) -> Marginals:
    name = f"{dataset.name()}_marginals"

    return Marginals(
        sp.Marginals(
            dataset=dataset.uuid(),
            name=name,
            statistics=statistics.protobuf()
            if statistics is not None
            else statistics,
            properties=properties,
        )
    )



import functools
import inspect
import logging
import typing as t

from cachetools import LRUCache
import pandas as pd

from sarus_data_spec.manager.typing import Computation, Manager
from sarus_data_spec.typing import Referrable

logger = logging.getLogger(__name__)


def retrieve_or_validate_cache(
    category: str, args: t.Tuple, kwargs: t.Dict
) -> t.Optional[LRUCache]:
    """
    Retrieves the cache for the specified category from the manager of the
    global context.

    Args:
        category (str): The category name for which the cache needs to be
        retrieved.

    Returns:
        LRUCache: The cache object associated with the given category.

    Raises:
        ValueError: If the category is not initialized in the cache manager.
        TypeError: If the retrieved cache is not an instance of LRUCache.
        NotImplementedError: If the global context manager's caches method
        is not implemented.
    """
    manager = get_manager_from_all_arguments(args, kwargs)

    try:
        caches = manager.caches()
    except NotImplementedError:
        return None

    if category not in caches:
        raise ValueError(
            f"""Category:{category} not initialized in the caching of
            the manager."""
        )

    cache = caches[category]
    if not isinstance(cache, LRUCache):
        raise TypeError(
            f"The cache for category '{category}' is not an LRUCache instance"
        )

    return cache


def get_manager_from_all_arguments(args: t.Tuple, kwargs: t.Dict) -> Manager:
    """Return one manager of at least one argument, otherwise
    return an error.
    """
    for arg in args:
        manager = get_manager_from_arg(arg)
        if manager is not None:
            return manager
    for _, v in kwargs.items():
        manager = get_manager_from_arg(v)
        if manager is not None:
            return manager

    raise ValueError("Impossible to retrieve a manager from the argument.")


def get_manager_from_arg(argument: t.Any) -> t.Optional[Manager]:
    """Return the manager of the argument if it exists"""
    if isinstance(argument, Referrable):
        referrable = t.cast(Referrable, argument)
        manager = referrable.manager()
        if not isinstance(manager, Manager):
            return None
        else:
            return manager
    else:
        return None


def process_argument(argument: t.Any) -> t.Any:
    """
    Processes an argument to extract a suitable representation for cache key
    construction.
    """
    if isinstance(argument, Referrable):
        referrable = t.cast(Referrable, argument)
        return referrable.uuid()
    elif isinstance(argument, Computation):
        return argument.task_name
    else:
        return argument


def build_cache_key(
    func: t.Callable, args: t.Tuple, kwargs: t.Dict, use_first_arg: bool
) -> t.Tuple:
    """
    Constructs a cache key based on the function's name and its arguments.

    Args:
        func (t.Callable): The function for which the cache key is being built.
        args (t.Tuple): The positional arguments passed to the function.
        kwargs (t.Dict): The keyword arguments passed to the function.
        use_first_arg (bool): Flag indicating whether to include the first
        argument in the key.
    """
    key: t.List[t.Any] = [func.__name__]
    start_index = 0 if use_first_arg else 1
    for arg in args[start_index:]:
        key.append(process_argument(arg))
    for k, v in kwargs.items():
        key.append((k, process_argument(v)))
    return tuple(key)


def safeguard_cache_integrity(result: t.Any) -> t.Any:
    """
    Creates a copy of the result if it's a mutable object (like a DataFrame
    or Series).
    This is to ensure that modifications to the returned object do not affect
    the cached value.
    """
    if isinstance(result, (pd.DataFrame, pd.Series)):
        return result.copy()
    return result


def lru_caching(
    category: str = "default", use_first_arg: bool = True
) -> t.Callable:
    """
    Decorator that provides caching functionality for functions, with support
    for
    both asynchronous and synchronous functions.

    Args:
        category (str): The cache category to use.
        use_first_arg (bool): Flag to include or exclude the first argument in
        the cache key.

    Returns:
        t.Callable: A decorator that when applied to a function, enables
        caching of its results.
    """

    def decorator(func: t.Callable) -> t.Any:
        is_async = inspect.iscoroutinefunction(func)

        @functools.wraps(func)
        async def async_wrapper(*args: t.Any, **kwargs: t.Any) -> t.Any:
            return await async_cache_logic(func, *args, **kwargs)

        @functools.wraps(func)
        def sync_wrapper(*args: t.Any, **kwargs: t.Any) -> t.Any:
            return sync_cache_logic(func, *args, **kwargs)

        async def async_cache_logic(
            func: t.Callable, *args: t.Any, **kwargs: t.Any
        ) -> t.Any:
            cache = retrieve_or_validate_cache(category, args, kwargs)
            if cache is None:
                return await func(*args, **kwargs)
            key = build_cache_key(func, args, kwargs, use_first_arg)

            # computing result
            if key in cache:
                logger.info(
                    f"Cache hit for for {func.__name__} with key: {key}"
                )
                result = cache[key]
            else:
                logger.info(
                    f"Computed and cached for {func.__name__} with key: {key}"
                )
                result = await func(*args, **kwargs)
                if result is not None:
                    cache[key] = result

            return safeguard_cache_integrity(result)

        def sync_cache_logic(
            func: t.Callable, *args: t.Any, **kwargs: t.Any
        ) -> t.Any:
            cache = retrieve_or_validate_cache(category, args, kwargs)
            if cache is None:
                return func(*args, **kwargs)
            key = build_cache_key(func, args, kwargs, use_first_arg)

            # computing result
            if key in cache:
                logger.info(
                    f"Cache hit for for {func.__name__} with key: {key}"
                )
                result = cache[key]
            else:
                logger.info(
                    f"Computed and cached for {func.__name__} with key: {key}"
                )
                result = func(*args, **kwargs)
                if result is not None:
                    cache[key] = result

            return safeguard_cache_integrity(result)

        return async_wrapper if is_async else sync_wrapper

    return decorator



import typing as t

from sarus_data_spec.manager.ops.base import ScalarImplementation


class PrivacyParams(ScalarImplementation):
    async def value(self) -> t.Any:
        assert self.scalar.is_privacy_params()
        points = self.scalar.protobuf().spec.privacy_params.points
        if len(points) != 1:
            raise NotImplementedError(
                "The PrivacyParams contains more than 1 point in the privacy "
                "profile."
            )
        return list(points)



from typing import Any, List, Literal, Optional, Sequence, Tuple, Union
import typing as t

import numpy as np
import pandas._typing as pdt

from sarus_data_spec.dataspec_validator.parameter_kind import DATASPEC, STATIC
from sarus_data_spec.dataspec_validator.signature import (
    SarusParameter,
    SarusSignature,
    SarusSignatureValue,
)

from .external_op import ExternalOpImplementation

try:
    from scipy.sparse import (
        bsr_array,
        bsr_matrix,
        coo_array,
        coo_matrix,
        csc_array,
        csc_matrix,
        csr_array,
        csr_matrix,
        dia_array,
        dia_matrix,
        dok_array,
        dok_matrix,
        lil_array,
        lil_matrix,
        spmatrix,
    )
except ModuleNotFoundError:
    bsr_matrix = Any
    coo_matrix = Any
    csc_matrix = Any
    csr_matrix = Any
    dok_matrix = Any
    lil_matrix = Any
    bsr_array = Any
    coo_matrix = Any
    coo_array = Any
    csc_array = Any
    csr_array = Any
    dia_matrix = Any
    dia_array = Any
    dok_array = Any
    lil_array = Any
    spmatrix = Any

Casting = Literal["no", "equiv", "safe", "same_kind", "unsafe"]
Order = Literal["K", "A", "C", "F"]


# ------ CONSTRUCTORS ------
class sp_bsr_matrix(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_BSR_MATRIX"
    _signature = SarusSignature(
        SarusParameter(
            name="arg1",
            annotation=Union[
                np.ndarray,
                bsr_matrix,
                Tuple[np.ndarray, Tuple[np.ndarray, np.ndarray]],
                Tuple[np.ndarray, np.ndarray, np.ndarray],
            ],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="shape",
            annotation=Optional[Tuple[int, int]],
            default=None,
        ),
        SarusParameter(
            name="dtype",
            annotation=Optional[Union[np.dtype, str]],
            default=None,
        ),
        SarusParameter(
            name="copy",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="blocksize",
            annotation=Optional[Tuple[int, int]],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> bsr_matrix:
        this, kwargs = signature.collect_kwargs_method()
        return bsr_matrix(this, **kwargs)


class sp_coo_matrix(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_COO_MATRIX"
    _signature = SarusSignature(
        SarusParameter(
            name="arg1",
            annotation=Union[
                np.ndarray,
                coo_matrix,
                Tuple[np.ndarray, Tuple[np.ndarray, np.ndarray]],
            ],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="shape",
            annotation=Optional[Tuple[int, int]],
            default=None,
        ),
        SarusParameter(
            name="dtype",
            annotation=Optional[np.dtype],
            default=None,
        ),
        SarusParameter(
            name="copy",
            annotation=Optional[bool],
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return coo_matrix(this, **kwargs)


class sp_csc_matrix(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_CSC_MATRIX"
    _signature = SarusSignature(
        SarusParameter(
            name="arg1",
            annotation=Union[
                Tuple[pdt.ArrayLike, Tuple[pdt.ArrayLike, pdt.ArrayLike]],
                Tuple[pdt.ArrayLike, pdt.ArrayLike, pdt.ArrayLike],
                pdt.ArrayLike,
                Tuple[int, int],
            ],
            default=None,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="shape",
            annotation=Optional[Tuple[int, int]],
            default=None,
        ),
        SarusParameter(
            name="dtype",
            annotation=Optional[type],
            default=None,
        ),
        SarusParameter(
            name="copy",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> csc_matrix:
        kwargs = signature.collect_kwargs()
        return csc_matrix(**kwargs)


class sp_csr_matrix(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_CSR_MATRIX"
    _signature = SarusSignature(
        SarusParameter(
            name="arg1",
            annotation=Union[
                Tuple[pdt.ArrayLike, Tuple[pdt.ArrayLike, pdt.ArrayLike]],
                Tuple[pdt.ArrayLike, pdt.ArrayLike, pdt.ArrayLike],
                pdt.ArrayLike,
                Tuple[int, int],
            ],
            default=None,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="shape",
            annotation=Optional[Tuple[int, int]],
            default=None,
        ),
        SarusParameter(
            name="dtype",
            annotation=Optional[type],
            default=None,
        ),
        SarusParameter(
            name="copy",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> csr_matrix:
        kwargs = signature.collect_kwargs()
        return csr_matrix(**kwargs)


class sp_dia_matrix(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_DIA_MATRIX"
    _signature = SarusSignature(
        SarusParameter(
            name="arg1",
            annotation=Union[
                Tuple[np.ndarray, np.ndarray],
                Tuple[int, int],
                dia_matrix,
                np.ndarray,
            ],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="shape",
            annotation=Optional[Tuple[int, int]],
            default=None,
        ),
        SarusParameter(
            name="dtype",
            annotation=Optional[Union[str, np.dtype]],
            default=None,
        ),
        SarusParameter(
            name="copy",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return dia_matrix(**kwargs)


class sp_dok_matrix(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_DOK_MATRIX"
    _signature = SarusSignature(
        SarusParameter(
            name="arg1",
            annotation=Union[np.ndarray, spmatrix],
            default=None,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="shape",
            annotation=Optional[Tuple[int, int]],
            default=None,
        ),
        SarusParameter(
            name="dtype",
            annotation=Optional[np.dtype],
            default=None,
        ),
        SarusParameter(
            name="copy",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return dok_matrix(**kwargs)


class sp_lil_matrix(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_LIL_MATRIX"
    _signature = SarusSignature(
        SarusParameter(
            name="arg1",
            annotation=Union[np.ndarray, spmatrix],
            default=None,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="shape",
            annotation=Optional[Tuple[int, int]],
            default=None,
        ),
        SarusParameter(
            name="dtype",
            annotation=Optional[np.dtype],
            default=None,
        ),
        SarusParameter(
            name="copy",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return lil_matrix(**kwargs)


class sp_bsr_array(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_BSR_ARRAY"
    _signature = SarusSignature(
        SarusParameter(
            name="arg1",
            annotation=Union[
                np.ndarray,
                spmatrix,
                Tuple[np.ndarray, Tuple[np.ndarray, np.ndarray]],
                Tuple[np.ndarray, np.ndarray, np.ndarray],
            ],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="shape",
            annotation=Optional[Tuple[int, int]],
            default=None,
        ),
        SarusParameter(
            name="dtype",
            annotation=Optional[Union[str, np.dtype]],
            default=None,
        ),
        SarusParameter(
            name="copy",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="blocksize",
            annotation=Optional[Tuple[int, int]],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> bsr_matrix:
        kwargs = signature.collect_kwargs()
        return bsr_array(**kwargs)


class sp_coo_array(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_COO_ARRAY"
    _signature = SarusSignature(
        SarusParameter(
            name="arg1",
            annotation=Union[
                np.ndarray,
                spmatrix,
                Tuple[np.ndarray, Tuple[np.ndarray, np.ndarray]],
            ],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="shape",
            annotation=Optional[Tuple[int, int]],
            default=None,
        ),
        SarusParameter(
            name="dtype",
            annotation=Optional[Union[str, np.dtype]],
            default=None,
        ),
        SarusParameter(
            name="copy",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> coo_matrix:
        kwargs = signature.collect_kwargs()
        return coo_array(**kwargs)


class sp_csc_array(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_CSC_ARRAY"
    _signature = SarusSignature(
        SarusParameter(
            name="arg1",
            annotation=Union[
                np.ndarray,
                spmatrix,
                Tuple[Union[np.ndarray, List], Tuple[np.ndarray, np.ndarray]],
                Tuple[np.ndarray, np.ndarray, np.ndarray],
            ],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="shape",
            annotation=Optional[Tuple[int, int]],
            default=None,
        ),
        SarusParameter(
            name="dtype",
            annotation=Optional[np.dtype],
            default=None,
        ),
        SarusParameter(
            name="copy",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return csc_array(**kwargs)


class sp_csr_array(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_CSR_ARRAY"
    _signature = SarusSignature(
        SarusParameter(
            name="arg1",
            annotation=Union[
                np.ndarray,
                spmatrix,
                Tuple[Union[np.ndarray, List], Tuple[np.ndarray, np.ndarray]],
                Tuple[np.ndarray, np.ndarray, np.ndarray],
            ],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="shape",
            annotation=Optional[Tuple[int, int]],
            default=None,
        ),
        SarusParameter(
            name="dtype",
            annotation=Optional[np.dtype],
            default=None,
        ),
        SarusParameter(
            name="copy",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return csr_array(**kwargs)


class sp_dia_array(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_DIA_ARRAY"
    _signature = SarusSignature(
        SarusParameter(
            name="arg1",
            annotation=Union[
                np.ndarray,
                dia_matrix,
                Tuple[Union[np.ndarray, Sequence[int]], Tuple[int, int]],
            ],
            default=None,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="shape",
            annotation=Optional[Tuple[int, int]],
            default=None,
        ),
        SarusParameter(
            name="dtype",
            annotation=Optional[np.dtype],
            default=None,
        ),
        SarusParameter(
            name="copy",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return dia_array(**kwargs)


class sp_dok_array(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_DOK_ARRAY"
    _signature = SarusSignature(
        SarusParameter(
            name="arg1",
            annotation=Union[np.ndarray, dok_matrix, Tuple[int, int]],
            default=None,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="shape",
            annotation=Optional[Tuple[int, int]],
            default=None,
        ),
        SarusParameter(
            name="dtype",
            annotation=Optional[np.dtype],
            default=None,
        ),
        SarusParameter(
            name="copy",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return dok_array(**kwargs)


class sp_lil_array(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_LIL_ARRAY"
    _signature = SarusSignature(
        SarusParameter(
            name="arg1",
            annotation=Union[np.ndarray, lil_matrix, Tuple[int, int]],
            default=None,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="shape",
            annotation=Optional[Tuple[int, int]],
            default=None,
        ),
        SarusParameter(
            name="dtype",
            annotation=Optional[np.dtype],
            default=None,
        ),
        SarusParameter(
            name="copy",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return lil_array(**kwargs)


# ------ Scipy sparse METHODS ------


class scipy_SCIPY_arcsin(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_ARCSIN"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=spmatrix,
            condition=DATASPEC | STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> spmatrix:
        this = signature.collect_args()[0]
        return this.arcsin()


class scipy_SCIPY_asfptype(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_ASFPTYPE"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=spmatrix,
            condition=DATASPEC | STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> spmatrix:
        this = signature.collect_args()[0]
        return this.asfptype()


class scipy_SCIPY_ceil(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_CEIL"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=spmatrix,
            condition=DATASPEC | STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> spmatrix:
        this = signature.collect_args()[0]
        return this.ceil()


class scipy_SCIPY_argmax(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_ARGMAX"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=spmatrix,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="axis",
            annotation=Any,
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Union[csr_matrix, int]:
        this, axis = signature.collect_args()
        return this.argmax(axis=axis)


class scipy_conj(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_CONJ"
    _signature = SarusSignature(
        SarusParameter(
            name="data",
            annotation=spmatrix,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="copy",
            annotation=bool,
            default=True,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> spmatrix:
        data, copy = signature.collect_args()
        return data.conj(copy=copy)


class scipy_conjugate(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_CONJUGATE"
    _signature = SarusSignature(
        SarusParameter(
            name="data",
            annotation=spmatrix,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="copy",
            annotation=bool,
            default=True,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        data, copy = signature.collect_args()
        return data.conjugate(copy=copy)


class scipy_count_nonzero(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_COUNT_NONZERO"
    _signature = SarusSignature(
        SarusParameter(
            name="data",
            annotation=spmatrix,
            condition=DATASPEC | STATIC,
        )
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (data,) = signature.collect_args()
        return data.count_nonzero()


class scipy_deg2rad(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_DEG2RAD"
    _signature = SarusSignature(
        SarusParameter(
            name="data",
            annotation=spmatrix,
            condition=DATASPEC | STATIC,
        )
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (data,) = signature.collect_args()
        return data.deg2rad()


class scipy_diagonal(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_DIAGONAL"
    _signature = SarusSignature(
        SarusParameter(
            name="data",
            annotation=spmatrix,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="k",
            annotation=int,
            default=0,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        data, k = signature.collect_args()
        return data.diagonal(k=k)


class scipy_dot(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_DOT"
    _signature = SarusSignature(
        SarusParameter(
            name="data",
            annotation=spmatrix,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="other",
            annotation=spmatrix,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> spmatrix:
        data, other = signature.collect_args()
        return data.dot(other)


class scipy_eliminate_zeros(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_ELIMINATE_ZEROS"
    _signature = SarusSignature(
        SarusParameter(
            name="data",
            annotation=spmatrix,
            condition=DATASPEC | STATIC,
        )
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (data,) = signature.collect_args()
        data.eliminate_zeros()
        return data


class scipy_expm1(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_EXPM1"
    _signature = SarusSignature(
        SarusParameter(
            name="data",
            annotation=spmatrix,
            condition=DATASPEC | STATIC,
        )
    )

    def call(self, signature: SarusSignatureValue) -> spmatrix:
        (data,) = signature.collect_args()
        return data.expm1()


class scipy_floor(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_FLOOR"
    _signature = SarusSignature(
        SarusParameter(
            name="data",
            annotation=spmatrix,
            condition=DATASPEC | STATIC,
        )
    )

    def call(self, signature: SarusSignatureValue) -> spmatrix:
        (data,) = signature.collect_args()
        return data.floor()


class scipy_log1p(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_LOG1P"
    _signature = SarusSignature(
        SarusParameter(
            name="data",
            annotation=spmatrix,
            condition=DATASPEC | STATIC,
        )
    )

    def call(self, signature: SarusSignatureValue) -> spmatrix:
        (data,) = signature.collect_args()
        return data.log1p()


class scipy_max(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_MAX"
    _signature = SarusSignature(
        SarusParameter(
            name="data",
            annotation=spmatrix,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="axis",
            annotation=Optional[Union[int, None]],
            default=None,
        ),
        SarusParameter(
            name="out",
            annotation=Optional[None],
            default=None,
        ),
    )

    def call(
        self, signature: SarusSignatureValue
    ) -> Union[spmatrix, int, float]:
        data, axis, out = signature.collect_args()
        return data.max(axis=axis, out=out)


class scipy_maximum(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_MAXIMUM"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=spmatrix,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="other",
            annotation=spmatrix,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this, other) = signature.collect_args()
        return this.maximum(other)


class scipy_mean(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_MEAN"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=spmatrix,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="axis",
            annotation=Optional[Union[int, None]],
            default=None,
        ),
        SarusParameter(
            name="dtype",
            annotation=Optional[Any],
            default=None,
        ),
        SarusParameter(
            name="out",
            annotation=Optional[np.matrix],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.mean(**kwargs)


class scipy_minimum(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_MINIMUM"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=spmatrix,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="other",
            annotation=spmatrix,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this, other) = signature.collect_args()
        return this.minimum(other)


class scipy_reshape(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_RESHAPE"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=spmatrix,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="shape",
            annotation=Tuple[int, int],
        ),
        SarusParameter(
            name="order",
            annotation=str,
            default="C",
        ),
        SarusParameter(
            name="copy",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.reshape(**kwargs)


class scipy_multiply(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_MULTIPLY"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=spmatrix,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="other",
            annotation=spmatrix,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> spmatrix:
        (this, other) = signature.collect_args()
        return this.multiply(other)


class scipy_nanmax(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_NANMAX"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=spmatrix,
            condition=DATASPEC | STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return this.nanmax()


class scipy_nanmin(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_NANMIN"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=spmatrix,
            condition=DATASPEC | STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return this.nanmin()


class scipy_nonzero(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_NONZERO"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=spmatrix,
            condition=DATASPEC | STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return this.nonzero()


class scipy_power(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_POWER"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=spmatrix,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(name="n", annotation=float),
        SarusParameter(
            name="dtype", annotation=t.Optional[type], default=None
        ),
    )

    def call(self, signature: SarusSignatureValue) -> spmatrix:
        this, n, dtype = signature.collect_args()
        return this.power(n, dtype=dtype)


class scipy_prune(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_PRUNE"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=spmatrix,
            condition=DATASPEC | STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return this.prune()


class scipy_rad2deg(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_RAD2DEG"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=spmatrix,
            condition=DATASPEC | STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> spmatrix:
        (this,) = signature.collect_args()
        return this.rad2deg()


class scipy_resize(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_RESIZE"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=spmatrix,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="shape",
            annotation=t.Tuple[int, int],
            condition=DATASPEC | STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> spmatrix:
        (this, shape) = signature.collect_args()
        return this.resize(*shape)


class scipy_rint(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_RINT"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=spmatrix,
            condition=DATASPEC | STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> spmatrix:
        (this,) = signature.collect_args()
        return this.rint()


class scipy_setdiag(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_SETDIAG"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=spmatrix,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="values",
            annotation=t.Union[t.Sequence, float, int],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="k",
            annotation=int,
            default=0,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> spmatrix:
        (this, values, k) = signature.collect_args()
        return this.setdiag(values, k)


class scipy_sign(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_SIGN"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=spmatrix,
            condition=DATASPEC | STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> spmatrix:
        (this,) = signature.collect_args()
        return this.sign()


class scipy_sin(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_SIN"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=spmatrix,
            condition=DATASPEC | STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> spmatrix:
        (this,) = signature.collect_args()
        return this.sin()


class scipy_sinh(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_SINH"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=spmatrix,
            condition=DATASPEC | STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> spmatrix:
        (this,) = signature.collect_args()
        return this.sinh()


class scipy_sort_indices(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_SORT_INDICES"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=spmatrix,
            condition=DATASPEC | STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        this.sort_indices()
        return this


class scipy_sorted_indices(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_SORTED_INDICES"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=spmatrix,
            condition=DATASPEC | STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> spmatrix:
        (this,) = signature.collect_args()
        return this.sorted_indices()


class scipy_sqrt(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_SQRT"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=spmatrix,
            condition=DATASPEC | STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> spmatrix:
        (this,) = signature.collect_args()
        return this.sqrt()


class scipy_sum(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_SUM"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=spmatrix,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="axis",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="dtype",
            annotation=Optional[t.Any],
            default=None,
        ),
        SarusParameter(
            name="out",
            annotation=Optional[np.matrix],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.sum(**kwargs)


class scipy_sum_duplicates(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_SUM_DUPLICATES"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=spmatrix,
            condition=DATASPEC | STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        this.sum_duplicates()
        return this


class scipy_tan(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_TAN"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=spmatrix,
            condition=DATASPEC | STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> spmatrix:
        (this,) = signature.collect_args()
        return this.tan()


class scipy_tanh(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_TANH"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=spmatrix,
            condition=DATASPEC | STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> spmatrix:
        (this,) = signature.collect_args()
        return this.tanh()


class scipy_trace(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_TRACE"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=spmatrix,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="offset",
            annotation=int,
            default=0,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this, offset) = signature.collect_args()
        return this.trace(offset)


class scipy_transpose(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_TRANSPOSE"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=spmatrix,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="axes",
            annotation=Optional[None],
            default=None,
        ),
        SarusParameter(
            name="copy",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.transpose(**kwargs)


class scipy_trunc(ExternalOpImplementation):
    _transform_id = "scipy.SCIPY_TRUNC"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=spmatrix,
            condition=DATASPEC | STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return this.trunc()



from __future__ import annotations

from datetime import datetime
from functools import partial
from typing import Callable, Mapping, Optional, Type, cast
import logging
import re
import typing as t

from sarus_data_spec import typing as st
from sarus_data_spec.base import Base, Referring
from sarus_data_spec.dataset import Dataset
from sarus_data_spec.manager.typing import Manager
from sarus_data_spec.protobuf.typing import ProtobufWithUUIDAndDatetime
import sarus_data_spec.protobuf as sp

logger = logging.getLogger(__name__)


class DataSpecErrorStatus(Exception):
    def __init__(self, relaunch_error_msg: t.Tuple[bool, str]) -> None:
        relaunch, error_msg = relaunch_error_msg
        self.relaunch = relaunch
        self.error_msg = error_msg

    def __str__(self) -> str:
        return self.error_msg

    def extract_last_exc(self) -> Optional[str]:
        """Extract from the traceback the part containing the last exception"""
        m = re.search(
            r"Traceback \(most recent call last\):\n(  File \".*\", line \d+, in .*\n  .*\n)+(([a-zA-Z_\.]+)(: (.*))?)$",  # noqa: E501
            self.error_msg.strip(),
        )

        return m.group(3) if m else None


class Status(Referring[sp.Status]):
    """A python class to describe status"""

    def __init__(self, protobuf: sp.Status, store: bool = True) -> None:
        self._referred = {
            protobuf.dataspec,
            protobuf.manager,
        }  # This has to be defined before it is initialized
        super().__init__(protobuf, store=store)

    def prototype(self) -> Type[sp.Status]:
        """Return the type of the underlying protobuf."""
        return sp.Status

    def datetime(self) -> datetime:
        return datetime.fromisoformat(self.protobuf().datetime)

    def update(
        self,
        task_stages: Optional[Mapping[str, st.Stage]] = None,
        properties: Optional[Mapping[str, str]] = None,
    ) -> t.Tuple[Status, bool]:
        """
        This method apply an atomic update
        to the last version of the status. The update
        is not performed if the transition from the old
        to the new stage is a regression (see method
        transitions_allowed)
        """
        # Now update the last version of this status
        new_status, is_updated = self.storage().update_referring_with_datetime(
            self.referred_uuid(),
            self.type_name(),
            partial(
                update_last, task_stages=task_stages, properties=properties
            ),
        )
        return t.cast(Status, new_status), is_updated

    def clear_task(self, task: str) -> t.Tuple[Status, bool]:
        """Creates a new status removing the task specified.
        To remove the task specified, the stage must be either
        ready or error (a static situation in which no one
        is processing the task).
        If the task does not exist, nothing is created"""
        # We copy the content of the current status

        def clear_update(
            status: st.Referring[ProtobufWithUUIDAndDatetime],
        ) -> t.Tuple[st.Referring[ProtobufWithUUIDAndDatetime], bool]:
            status = t.cast(Status, status)
            proto = status.protobuf()
            # We update its timestamp and properties
            proto.datetime = datetime.now().isoformat()
            stage = proto.task_stages.pop(task)
            if stage.WhichOneof("stage") is None:
                return status, False

            # we can create a new empty status only
            # if the current stage is static ie ready
            # (for example when removing cache) or
            # error (relaunch after timeout)
            else:
                can_be_cleared = transition_allowed(
                    existing_stage=stage, new_stage=None, task_name=task
                )
                if can_be_cleared:
                    # create a new data spec object
                    return Status(proto, store=False), True
                return status, False

        new_status, is_updated = self.storage().update_referring_with_datetime(
            self.referred_uuid(), self.type_name(), clear_update
        )
        return cast(Status, new_status), is_updated

    def task(self, task: str) -> t.Optional[Stage]:
        stage = self.protobuf().task_stages.get(task, None)
        return Stage(stage) if stage is not None else stage

    def task_stages(self) -> t.Mapping[str, Stage]:
        return {
            stage_name: Stage(stage_proto)
            for stage_name, stage_proto in self.protobuf().task_stages.items()
        }

    def pending(
        self,
    ) -> bool:
        """this is true if all tasks have status at least pending
        (i.e. pending, processing or ready) and at least one is pending"""
        has_one = False
        all_better = True
        task_stages = self.protobuf().task_stages
        if task_stages is not None:
            for task in task_stages:
                has_one = has_one or task_stages[task].HasField("pending")
                all_better = all_better and not (
                    task_stages[task].HasField("error")
                )
        return has_one and all_better

    def processing(
        self,
    ) -> bool:
        """this is true if all tasks have status at least processing
        (i.e. processing or ready) and at least one is processing"""
        has_one = False
        all_better = True
        task_stages = self.protobuf().task_stages
        if task_stages is not None:
            for task in task_stages:
                has_one = has_one or task_stages[task].HasField("processing")
                all_better = all_better and not (
                    task_stages[task].HasField("error")
                    or task_stages[task].HasField("pending")
                )
        return has_one and all_better

    def ready(
        self,
    ) -> bool:  # TODO this should be true if all tasks have status ready
        """this is true if all tasks have status ready"""
        all_ready = True
        task_stages = self.protobuf().task_stages
        if task_stages is not None:
            for task in task_stages:
                all_ready = all_ready and task_stages[task].HasField("ready")
        return all_ready

    def error(
        self,
    ) -> bool:  # TODO this should be true if any task have status error
        """this is true if any tasks have status error"""
        has_error = False
        task_stages = self.protobuf().task_stages
        if task_stages is not None:
            for task in task_stages:
                has_error = has_error or task_stages[task].HasField("error")
        return has_error

    def dataset(self) -> Dataset:
        dataspec = self.dataspec()
        assert isinstance(dataspec, Dataset)
        return dataspec

    def dataspec(self) -> st.DataSpec:
        return cast(
            st.DataSpec, self.storage().referrable(self._protobuf.dataspec)
        )

    def owner(
        self,
    ) -> Manager:
        # TODO: Maybe find a better name, but this was shadowing
        # the actual manager of this object.  # noqa: E501
        return cast(Manager, self.storage().referrable(self._protobuf.manager))


def transition_allowed(
    existing_stage: t.Optional[sp.Status.Stage],
    new_stage: t.Optional[sp.Status.Stage],
    task_name: str,
) -> bool:
    """Method that verifies that the new stage for a given task
    can be set to the existing. The rule is that stages can only
    go towards a more advanced stage and never backwards, except
    for error and ready that can be reset to None.
    When the stage is None, it means that the task is not present"""

    if existing_stage is None:
        return True  # all transitions allowed if no task exists

    allowed_transitions = {
        "pending": ["processing", "ready", "error"],
        "processing": ["ready", "error"],
        "ready": ["error", None],
        "error": [None, "ready"],
        None: ["pending", "processing", "ready", "error"],
        # the transition error to ready can occur
        # if a long task is processed, meanwhile another worker sets an error
        # because it timeouts after waiting, then the first worker finishes
    }
    if new_stage is not None:
        if (
            new_stage.WhichOneof("stage")  # type:ignore
            in allowed_transitions[existing_stage.WhichOneof("stage")]
        ):
            return True
        else:
            logger.warning(
                f"Trying to update a status with stage "
                f"{new_stage.WhichOneof('stage')} for the task {task_name} "
                f"while the existing stage "
                f"is {existing_stage.WhichOneof('stage')}, "
                f"therefore the update will be ignored"
            )
        return False

    else:
        if (
            new_stage  # type: ignore
            in allowed_transitions[existing_stage.WhichOneof("stage")]
        ):
            return True

        else:
            logger.warning(
                f"Trying to update a status with stage "
                f"{new_stage} for the task {task_name} "
                f"while the existing stage "
                f"is {existing_stage.WhichOneof('stage')}, "
                f"therefore the update will be ignored"
            )
            return False


# Builders
def status(
    dataspec: st.DataSpec,
    task_stages: Optional[Mapping[str, st.Stage]] = None,
    properties: Optional[Mapping[str, str]] = None,
    manager: Optional[Manager] = None,
) -> t.Tuple[st.Status, bool]:
    """A builder to ease the construction of a status
    - dataspec: is the dataspec, the status is added to
    - manager: is which manager, if None,
        the default manager of the dataspec is used
    """
    # Use the right manager
    manager = dataspec.manager() if manager is None else manager
    # If a status already exists for the dataspec
    # and manager simply create an updated version of it
    last = last_status(dataspec=dataspec, manager=manager)
    if last is None:
        status = Status(
            sp.Status(
                dataspec=dataspec.uuid(),
                manager=manager.uuid(),
                datetime=datetime.now().isoformat(),
                task_stages=None
                if task_stages is None
                else {
                    task: task_stages[task].protobuf() for task in task_stages
                },
                properties=properties,
            ),
            store=False,
        )
        (
            new_status,
            is_updated,
        ) = status.storage().create_referring_with_datetime(
            status,
            partial(
                update_last, task_stages=task_stages, properties=properties
            ),
        )
        return t.cast(st.Status, new_status), is_updated
    else:
        return last.update(task_stages=task_stages, properties=properties)


def last_status(
    dataspec: st.DataSpec,
    manager: Optional[Manager] = None,
    task: t.Optional[str] = None,
) -> Optional[st.Status]:
    """Return a DataSpec's last status by sorted datetime."""
    manager = dataspec.manager() if manager is None else manager
    status = cast(
        Optional[st.Status],
        manager.storage().last_referring(
            (dataspec.uuid(), manager.uuid()), sp.type_name(sp.Status)
        ),
    )

    if status is None or task is None:
        return status

    if status.task(task=task) is None:
        return None

    return status


class Stage(Base[sp.Status.Stage]):
    """A simple wrapper type to simplify protobuf usage"""

    def accept(self, visitor: st.StageVisitor) -> None:
        dispatch: Callable[[], None] = {
            "pending": visitor.pending,
            "processing": visitor.processing,
            "ready": visitor.ready,
            "error": visitor.error,
        }[cast(str, self.protobuf().WhichOneof("stage"))]
        dispatch()

    def stage(self) -> str:
        return cast(str, self.protobuf().WhichOneof("stage"))

    def ready(self) -> bool:
        return self.stage() == "ready"

    def pending(self) -> bool:
        return self.stage() == "pending"

    def processing(self) -> bool:
        return self.stage() == "processing"

    def error(self) -> bool:
        return self.stage() == "error"


# Builders for stages
def pending_stage(properties: Optional[Mapping[str, str]] = None) -> Stage:
    return Stage(
        sp.Status.Stage(
            pending=sp.Status.Stage.Pending(), properties=properties
        )
    )


def processing_stage(properties: Optional[Mapping[str, str]] = None) -> Stage:
    return Stage(
        sp.Status.Stage(
            processing=sp.Status.Stage.Processing(), properties=properties
        )
    )


def ready_stage(properties: Optional[Mapping[str, str]] = None) -> Stage:
    return Stage(
        sp.Status.Stage(ready=sp.Status.Stage.Ready(), properties=properties)
    )


def error_stage(properties: Optional[Mapping[str, str]] = None) -> Stage:
    return Stage(
        sp.Status.Stage(error=sp.Status.Stage.Error(), properties=properties)
    )


# Builders for statuses


def properties(
    dataspec: st.DataSpec,
    properties: Optional[Mapping[str, str]],
    manager: Optional[Manager] = None,
) -> t.Tuple[st.Status, bool]:
    return status(
        dataspec,
        task_stages=None,
        properties=properties,
        manager=manager,
    )


def pending(
    dataspec: st.DataSpec,
    task: str,
    properties: Optional[Mapping[str, str]] = None,
    manager: Optional[Manager] = None,
) -> t.Tuple[st.Status, bool]:
    return status(
        dataspec,
        task_stages={task: pending_stage(properties=properties)},
        properties=None,
        manager=manager,
    )


def processing(
    dataspec: st.DataSpec,
    task: str,
    properties: Optional[Mapping[str, str]] = None,
    manager: Optional[Manager] = None,
) -> t.Tuple[st.Status, bool]:
    return status(
        dataspec,
        task_stages={task: processing_stage(properties=properties)},
        properties=None,
        manager=manager,
    )


def ready(
    dataspec: st.DataSpec,
    task: str,
    properties: Optional[Mapping[str, str]] = None,
    manager: Optional[Manager] = None,
) -> t.Tuple[st.Status, bool]:
    return status(
        dataspec,
        task_stages={task: ready_stage(properties=properties)},
        properties=None,
        manager=manager,
    )


def error(
    dataspec: st.DataSpec,
    task: str,
    properties: Optional[Mapping[str, str]] = None,
    manager: Optional[Manager] = None,
) -> t.Tuple[st.Status, bool]:
    return status(
        dataspec,
        task_stages={task: error_stage(properties=properties)},
        properties=None,
        manager=manager,
    )


def update_last(
    status: st.Referring[ProtobufWithUUIDAndDatetime],
    task_stages: Optional[Mapping[str, st.Stage]],
    properties: Optional[Mapping[str, str]],
) -> t.Tuple[st.Referring[ProtobufWithUUIDAndDatetime], bool]:
    """Returns whether the update should be done, ie transitions
    are respected and the status that can be updated in case"""
    # We copy the content of the current status
    proto = cast(Status, status).protobuf()
    # We update its timestamp and properties
    proto.datetime = datetime.now().isoformat()
    if properties is not None:
        proto.properties.update(properties)
        # Protobuf implementation of the merge
        # does not work as expected on maps...
    if task_stages is not None:
        for task in task_stages:
            stage = task_stages[task].protobuf()
            if (
                task in proto.task_stages
                and proto.task_stages[task].WhichOneof("stage")
                == task_stages[task].protobuf().WhichOneof("stage")
                and not stage.properties == proto.task_stages[task].properties
            ):
                proto.task_stages[task].properties.update(stage.properties)
            else:
                if transition_allowed(
                    existing_stage=proto.task_stages.get(task, None),
                    new_stage=stage,
                    task_name=task,
                ):
                    proto.task_stages[task].CopyFrom(stage)
                else:
                    return status, False
    # And return a new UNSTORED data spec object
    return Status(proto, store=False), True


def status_error_policy(stage: st.Stage) -> bool:
    """This methods returns whether a given error message
    should be reset to None or not. Currently, the quick
    shortcut is to check whether the error message starts
    with a TimeoutError, in this case, we want to retry"""
    assert stage.error()
    return stage.properties().get("relaunch", "False") == str(True)


def error_aggregation(errors: t.List[Exception]) -> Exception:
    """Takes as input a list of exceptions, the first error
    that is not a DataSpecErrorStatus or that has not a relaunch.
    If they are all DataSpecErrorStatuses with relaunch, returns
    the first"""

    for error in errors:
        if isinstance(error, DataSpecErrorStatus) and error.relaunch:
            continue
        else:
            return error

    return errors[0]



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import sarus_data_spec.protobuf.path_pb2
import sarus_data_spec.protobuf.statistics_pb2
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor = ...

class Links(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    class PropertiesEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: typing.Text = ...
        value: typing.Text = ...
        def __init__(self,
            *,
            key : typing.Text = ...,
            value : typing.Text = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...

    class LinkStat(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        class PropertiesEntry(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
            KEY_FIELD_NUMBER: builtins.int
            VALUE_FIELD_NUMBER: builtins.int
            key: typing.Text = ...
            value: typing.Text = ...
            def __init__(self,
                *,
                key : typing.Text = ...,
                value : typing.Text = ...,
                ) -> None: ...
            def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...

        POINTING_FIELD_NUMBER: builtins.int
        POINTED_FIELD_NUMBER: builtins.int
        DISTRIBUTION_FIELD_NUMBER: builtins.int
        PROPERTIES_FIELD_NUMBER: builtins.int
        @property
        def pointing(self) -> sarus_data_spec.protobuf.path_pb2.Path: ...
        @property
        def pointed(self) -> sarus_data_spec.protobuf.path_pb2.Path: ...
        @property
        def distribution(self) -> sarus_data_spec.protobuf.statistics_pb2.Distribution: ...
        @property
        def properties(self) -> google.protobuf.internal.containers.ScalarMap[typing.Text, typing.Text]:
            """Needed for compatibility with typing"""
            pass
        def __init__(self,
            *,
            pointing : typing.Optional[sarus_data_spec.protobuf.path_pb2.Path] = ...,
            pointed : typing.Optional[sarus_data_spec.protobuf.path_pb2.Path] = ...,
            distribution : typing.Optional[sarus_data_spec.protobuf.statistics_pb2.Distribution] = ...,
            properties : typing.Optional[typing.Mapping[typing.Text, typing.Text]] = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["distribution",b"distribution","pointed",b"pointed","pointing",b"pointing"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["distribution",b"distribution","pointed",b"pointed","pointing",b"pointing","properties",b"properties"]) -> None: ...

    UUID_FIELD_NUMBER: builtins.int
    DATASET_FIELD_NUMBER: builtins.int
    LINKS_STATISTICS_FIELD_NUMBER: builtins.int
    PROPERTIES_FIELD_NUMBER: builtins.int
    uuid: typing.Text = ...
    dataset: typing.Text = ...
    """uuid"""

    @property
    def links_statistics(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Links.LinkStat]: ...
    @property
    def properties(self) -> google.protobuf.internal.containers.ScalarMap[typing.Text, typing.Text]: ...
    def __init__(self,
        *,
        uuid : typing.Text = ...,
        dataset : typing.Text = ...,
        links_statistics : typing.Optional[typing.Iterable[global___Links.LinkStat]] = ...,
        properties : typing.Optional[typing.Mapping[typing.Text, typing.Text]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["dataset",b"dataset","links_statistics",b"links_statistics","properties",b"properties","uuid",b"uuid"]) -> None: ...
global___Links = Links



from typing import Collection, List, Optional, Set, Union
import logging
import typing as t

try:
    from sarus_differential_privacy.query import PrivateQuery
except ModuleNotFoundError as e_groupby_dp:
    if "sarus_differential_privacy" not in str(e_groupby_dp):
        raise
    PrivateQuery = t.Any  # type: ignore

from sarus_data_spec.dataspec_validator.typing import DataspecValidator
import sarus_data_spec.typing as st

logger = logging.getLogger(__name__)


def find_dataspec_from_constraint(
    dataspec_validator: DataspecValidator,
    dataspec: st.DataSpec,
    kind: st.ConstraintKind,
    public_context: Collection[str],
    privacy_limit: Optional[st.PrivacyLimit],
    salt: Optional[int] = None,
) -> Optional[st.DataSpec]:
    # Current dataspec verifies the constraint?
    for constraint in dataspec_validator.verified_constraints(dataspec):
        if dataspec_validator.verifies(
            variant_constraint=constraint,
            kind=kind,
            public_context=public_context,
            privacy_limit=privacy_limit,
            salt=salt,
        ):
            return dataspec

    # Current dataspec has a variant that verifies the constraint?
    for variant in dataspec.variants():
        for constraint in dataspec_validator.verified_constraints(variant):
            if dataspec_validator.verifies(
                variant_constraint=constraint,
                kind=kind,
                public_context=public_context,
                privacy_limit=privacy_limit,
                salt=salt,
            ):
                return variant

    return None


def graph_private_queries(dataspec: st.DataSpec) -> List[PrivateQuery]:
    """Return all the private queries from a dataspec to the sources."""

    class PrivateQueriesVisitor(st.Visitor, st.TransformVisitor):
        visited: Set[Union[st.DataSpec, st.Transform]] = set()
        private_queries: List[PrivateQuery] = list()

        def all(self, visited: Union[st.DataSpec, st.Transform]) -> None:
            pass

        def transformed(
            self,
            visited: st.DataSpec,
            transform: st.Transform,
            *arguments: st.DataSpec,
            **named_arguments: st.DataSpec,
        ) -> None:
            if visited not in self.visited:
                self.private_queries.extend(visited.private_queries())

                for argument in arguments:
                    argument.accept(self)
                for _, argument in named_arguments.items():
                    argument.accept(self)
                self.visited.add(visited)

        def other(self, visited: Union[st.DataSpec, st.Transform]) -> None:
            if visited not in self.visited:
                self.visited.add(visited)

        def composed(
            self,
            visited: st.Transform,
            transform: st.Transform,
            *arguments: st.Transform,
            **named_arguments: st.Transform,
        ) -> None:
            for argument in arguments:
                argument.accept(self)
            for _, argument in named_arguments.items():
                argument.accept(self)

        def variable(
            self,
            visited: st.Transform,
            name: str,
            position: int,
        ) -> None:
            pass

    visitor = PrivateQueriesVisitor()
    dataspec.accept(visitor)
    private_queries = visitor.private_queries
    return private_queries



import logging
import typing as t
import warnings

try:
    from sarus_llm.torch_interface.sample import TextSampler
    from sarus_llm.torch_interface.typing import ModelProvider
except (ModuleNotFoundError, ValueError, RuntimeError):
    warnings.warn("Sarus LLM not available")

import pyarrow as pa

from sarus_data_spec.constants import PUBLIC
from sarus_data_spec.dataset import Dataset
from sarus_data_spec.manager.ops.processor.standard.standard_op import (  # noqa: E501
    StandardDatasetImplementation,
    StandardDatasetStaticChecker,
)
from sarus_data_spec.scalar import Scalar
from sarus_data_spec.schema import schema
from sarus_data_spec.type import Struct, Text
import sarus_data_spec.protobuf as sp
import sarus_data_spec.typing as st

logger = logging.getLogger()


class GenerateFromModelStaticChecker(StandardDatasetStaticChecker):
    async def schema(self) -> st.Schema:
        _, kwargs = self.dataset.parents()
        pretrained_model = kwargs["model"]
        assert isinstance(pretrained_model, Scalar)
        prompts = kwargs["prompts"]
        assert isinstance(prompts, Dataset)
        is_public = pretrained_model.is_public() and prompts.is_public()
        return schema(
            dataset=self.dataset,
            schema_type=Struct(
                {"text": Text()}, properties={PUBLIC: str(is_public)}
            ),
        )


class GenerateFromModel(StandardDatasetImplementation):
    async def to_arrow(
        self, batch_size: int
    ) -> t.AsyncIterator[pa.RecordBatch]:
        _, kwargs = self.dataset.parents()
        pretrained_model = kwargs["model"]
        prompts = kwargs["prompts"]
        transform = self.dataset.transform()

        assert (
            pretrained_model.prototype() == sp.Scalar  # type: ignore # noqa: E501
        ), "pretrained model should be a Scalar"
        pretrained_ds = t.cast(st.Scalar, pretrained_model)
        model_provider = t.cast(
            ModelProvider, await pretrained_ds.async_value()
        )

        assert prompts.prototype() == sp.Dataset, "prompts should be a Dataset"
        prompts_ds = t.cast(st.Dataset, prompts)

        specs = transform.protobuf().spec.generate_from_model
        sampler = TextSampler(
            model_provider=model_provider,
            quantize=True,
            temperature=specs.temperature,
            max_new_tokens=specs.max_new_tokens,
        )

        async def async_iterator() -> t.AsyncIterator[pa.RecordBatch]:
            async for batch in await prompts_ds.async_to_arrow(batch_size):
                text_samples = pa.array(
                    sampler.generate(prompts=batch.to_pydict()["prompt"])
                )
                yield pa.RecordBatch.from_struct_array(
                    pa.StructArray.from_arrays(
                        arrays=[text_samples], names=["text"]
                    )
                )

        return async_iterator()



import logging
import typing as t
import warnings

import numpy as np
import pyarrow as pa

from sarus_data_spec.arrow.array import convert_record_batch
from sarus_data_spec.constants import DATA, DATASET_SLUGNAME
from sarus_data_spec.dataset import Dataset
from sarus_data_spec.manager.ops.processor.standard.sample import fast_gather
from sarus_data_spec.manager.ops.processor.standard.standard_op import (  # noqa: E501
    StandardDatasetImplementation,
    StandardDatasetStaticChecker,
)
from sarus_data_spec.path import straight_path
from sarus_data_spec.scalar import Scalar
from sarus_data_spec.schema import schema
import sarus_data_spec.typing as st

try:
    import sqlalchemy as sa
except ModuleNotFoundError:
    warnings.warn("sqlalchemy not installed. No sampling on bigdata")

try:
    from sarus_data_spec.manager.ops.sql_utils.bigdata import (
        path_to_quoted_string,
        sqlalchemy_query_to_string,
    )
    from sarus_data_spec.manager.ops.sql_utils.queries import nest_queries
    from sarus_data_spec.manager.ops.sql_utils.schema_translations import (
        async_sa_metadata_from_dataset,
    )
except ModuleNotFoundError:
    warnings.warn("sql utils not installed.")

logger = logging.getLogger(__name__)


class ToSmallDataStaticChecker(StandardDatasetStaticChecker):
    async def schema(self) -> st.Schema:
        parent_schema = await self.parent_schema()
        return schema(
            self.dataset,
            schema_type=parent_schema.type(),
            privacy_unit_tracking_paths=parent_schema.protobuf().privacy_unit,
            properties=parent_schema.properties(),
            name=self.dataset.properties().get(DATASET_SLUGNAME, None),
        )


class ToSmallData(StandardDatasetImplementation):
    async def size(self) -> st.Size:
        raise NotImplementedError

    async def multiplicity(self) -> st.Multiplicity:
        raise NotImplementedError

    async def marginals(self) -> st.Marginals:
        raise NotImplementedError

    async def to_arrow(
        self, batch_size: int
    ) -> t.AsyncIterator[pa.RecordBatch]:
        parent = t.cast(Dataset, self.parent())
        if parent.manager().is_big_data(parent):
            queries = await self.sql_implementation()
            assert queries is not None
            nested_queries = nest_queries(queries)
            schema = await self.dataset.async_schema()
            return await self.dataset.manager().async_sql(
                parent,
                query=nested_queries,
                batch_size=batch_size,
                result_type=schema.type(),
            )
        return await self._arrow_to_arrow(batch_size)

    async def _arrow_to_arrow(
        self, batch_size: int
    ) -> t.AsyncIterator[pa.RecordBatch]:
        previous_schema = await self.parent_schema()
        parent_batches = [
            convert_record_batch(batch, previous_schema.type())
            async for batch in await self.parent_to_arrow(
                batch_size=batch_size
            )
        ]
        struct_arr = pa.concat_arrays(parent_batches)
        # maybe to add
        # if len(previous_schema.tables()) == 1:
        #    # to change here
        #    return await limit_arrow_to_arrow(self, batch_size)

        size_dict = await to_small_data_sizes(self.dataset)
        to_small_data_proto = (
            self.dataset.transform().protobuf().spec.to_small_data
        )
        random_sampling = to_small_data_proto.random_sampling
        seed = Scalar(to_small_data_proto.seed).value()
        generator = np.random.default_rng(seed)

        if struct_arr.type.get_field_index(DATA) == -1:
            array_to_take = struct_arr
        else:
            array_to_take = struct_arr.flatten()[
                struct_arr.type.get_field_index(DATA)
            ]

        indices_to_take = limit_indices_from_array(
            array=array_to_take,
            _type=previous_schema.data_type(),
            selected_indices=pa.array(
                np.linspace(0, len(struct_arr) - 1, len(struct_arr), dtype=int)
            ),
            new_size_dict=size_dict,
            curr_path=[DATA],
            random_sampling=random_sampling,
            random_gen=generator,
        )
        return fast_gather(
            indices=indices_to_take.to_numpy(),
            batches=pa.Table.from_arrays(
                struct_arr.flatten(),
                names=[one_child.name for one_child in list(struct_arr.type)],
            ).to_batches(max_chunksize=1000),
            batch_size=batch_size,
        )

    async def sql_implementation(
        self,
    ) -> t.Optional[t.Dict[t.Tuple[str, ...], str]]:
        schema = await self.parent_schema()
        to_small_data_proto = (
            self.dataset.transform().protobuf().spec.to_small_data
        )
        random_sampling = to_small_data_proto.random_sampling
        size_dict = await to_small_data_sizes(self.dataset)
        sqlalchemy_metadata = await async_sa_metadata_from_dataset(
            self.dataset
        )
        queries = {}
        schema_name = schema.name()
        for table_path in schema.tables():
            size_table = size_dict[table_path]
            curr_path = table_path.to_strings_list()[0]
            full_tablename = (
                curr_path[1:] if (curr_path[0] == DATA) else curr_path
            )
            is_there_table_name = len(full_tablename) > 0
            if not is_there_table_name:
                full_tablename = [schema_name]
            sa_table_name = path_to_quoted_string(
                straight_path(full_tablename)
            )
            sa_table = sqlalchemy_metadata.tables[sa_table_name]
            if random_sampling:
                sample_query = (
                    sa.select(sa_table)
                    .order_by(sa.func.random())
                    .limit(size_table)
                )
            else:
                sample_query = sa.select(sa_table).limit(size_table)
            if not is_there_table_name:
                full_tablename = [""]
            queries[tuple(full_tablename)] = sqlalchemy_query_to_string(
                sample_query
            )
        return queries


async def to_small_data_sizes(
    dataset: st.Dataset,
) -> t.Dict[st.Path, int]:
    """Get the sampling rates for each table"""
    parent_ds = t.cast(Dataset, dataset.parents()[0][0])
    to_small_data_proto = dataset.transform().protobuf().spec.to_small_data
    sampled_size = to_small_data_proto.size
    schema = await parent_ds.manager().async_schema(parent_ds)
    table_paths = schema.tables()
    size_per_table = int(int(sampled_size) / len(table_paths))

    # test if min_table_size is not too high
    return {table_path: size_per_table for table_path in table_paths}


def limit_indices_from_array(
    array: pa.Array,
    _type: st.Type,
    selected_indices: pa.Array,
    new_size_dict: t.Dict[st.Path, int],
    curr_path: t.List[str],
    random_gen: np.random.Generator,
    random_sampling: bool,
) -> pa.Array:
    class IndicesLimit(st.TypeVisitor):
        indices: pa.Array = selected_indices
        batch_array: pa.Array = array

        def Union(
            self,
            fields: t.Mapping[str, st.Type],
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            new_indices = []
            for item_name, item_type in fields.items():
                index_element = self.batch_array.type.get_field_index(
                    item_name
                )
                filter_idex = pa.array(
                    np.equal(
                        self.batch_array.field("field_selected"),
                        np.array(item_name),
                    )
                )
                updated_indices = limit_indices_from_array(
                    self.batch_array.filter(filter_idex).flatten()[
                        index_element
                    ],
                    item_type,
                    self.indices.filter(filter_idex),
                    new_size_dict,
                    [*(el for el in curr_path), item_name],
                    random_gen,
                    random_sampling,
                )
                new_indices.append(updated_indices)
            self.indices = pa.concat_arrays(new_indices)

        def Struct(
            self,
            fields: t.Mapping[str, st.Type],
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            new_size = new_size_dict[straight_path(curr_path.copy())]
            # taking the first new_size indices
            if random_sampling:
                self.indices = pa.array(
                    random_gen.choice(
                        self.indices,
                        replace=False,
                        size=min(new_size, len(self.indices), len(array)),
                    )
                )
            else:
                self.indices = self.indices.slice(0, new_size)

        def Id(
            self,
            unique: bool,
            reference: t.Optional[st.Path] = None,
            base: t.Optional[st.IdBase] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError

        def Optional(
            self,
            type: st.Type,
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError

        def Null(
            self, properties: t.Optional[t.Mapping[str, str]] = None
        ) -> None:
            raise NotImplementedError

        def Unit(
            self, properties: t.Optional[t.Mapping[str, str]] = None
        ) -> None:
            raise NotImplementedError

        def Boolean(
            self, properties: t.Optional[t.Mapping[str, str]] = None
        ) -> None:
            raise NotImplementedError

        def Integer(
            self,
            min: int,
            max: int,
            base: st.IntegerBase,
            possible_values: t.Iterable[int],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError

        def Enum(
            self,
            name: str,
            name_values: t.Sequence[t.Tuple[str, int]],
            ordered: bool,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError

        def Float(
            self,
            min: float,
            max: float,
            base: st.FloatBase,
            possible_values: t.Iterable[float],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError

        def Text(
            self,
            encoding: str,
            possible_values: t.Iterable[str],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError

        def Bytes(
            self, properties: t.Optional[t.Mapping[str, str]] = None
        ) -> None:
            raise NotImplementedError

        def List(
            self,
            type: st.Type,
            max_size: int,
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError

        def Array(
            self,
            type: st.Type,
            shape: t.Tuple[int, ...],
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError

        def Datetime(
            self,
            format: str,
            min: str,
            max: str,
            base: st.DatetimeBase,
            possible_values: t.Iterable[str],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError

        def Time(
            self,
            format: str,
            min: str,
            max: str,
            base: st.TimeBase,
            possible_values: t.Iterable[str],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError

        def Date(
            self,
            format: str,
            min: str,
            max: str,
            base: st.DateBase,
            possible_values: t.Iterable[str],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError

        def Duration(
            self,
            unit: str,
            min: int,
            max: int,
            possible_values: t.Iterable[int],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError

        def Constrained(
            self,
            type: st.Type,
            constraint: st.Predicate,
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError

        def Hypothesis(
            self,
            *types: t.Tuple[st.Type, float],
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError

    visitor = IndicesLimit()
    _type.accept(visitor)
    return visitor.indices



from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numpy as np
import numpy.typing as npt
import pandas as pd

from sarus_data_spec.dataspec_validator.signature import (
    SarusParameter,
    SarusSignature,
    SarusSignatureValue,
)

from ..external_op import ExternalOpImplementation

try:
    from sklearn import inspection
    from sklearn.base import BaseEstimator
except ModuleNotFoundError:
    BaseEstimator = Any


class sk_permutation_importance(ExternalOpImplementation):
    _transform_id = "sklearn.SK_PERMUTATION_IMPORTANCE"
    _signature = SarusSignature(
        SarusParameter(
            name="estimator",
            annotation=BaseEstimator,
        ),
        SarusParameter(
            name="X",
            annotation=Union[np.ndarray, pd.DataFrame],
        ),
        SarusParameter(
            name="y",
            annotation=Optional[npt.ArrayLike],
        ),
        SarusParameter(
            name="scoring",
            annotation=Optional[
                Union[
                    str,
                    Callable,
                    List,
                    Tuple,
                    Dict,
                ]
            ],
            default=None,
        ),
        SarusParameter(
            name="n_repeats",
            annotation=int,
            default=5,
        ),
        SarusParameter(
            name="n_jobs",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="random_state",
            annotation=Optional[Union[int, np.random.RandomState]],
            default=None,
        ),
        SarusParameter(
            name="sample_weight",
            annotation=Optional[npt.ArrayLike],
            default=None,
        ),
        SarusParameter(
            name="max_samples",
            annotation=Union[int, float],
            default=1.0,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return inspection.permutation_importance(**kwargs)
